import json
from unittest.mock import patch

from django.test import TestCase
from draftjs_exporter.dom import DOM
from draftjs_exporter.html import HTML as HTMLExporter

from wagtail.admin.rich_text.converters.contentstate import (
    ContentstateConverter,
    persist_key_for_block,
)
from wagtail.embeds.models import Embed


def content_state_equal(v1, v2, match_keys=False):
    "Test whether two contentState structures are equal, ignoring 'key' properties if match_keys=False"
    if type(v1) != type(v2):
        return False

    if isinstance(v1, dict):
        if set(v1.keys()) != set(v2.keys()):
            return False
        return all(
            (k == "key" and not match_keys)
            or content_state_equal(v, v2[k], match_keys=match_keys)
            for k, v in v1.items()
        )
    elif isinstance(v1, list):
        if len(v1) != len(v2):
            return False
        return all(
            content_state_equal(a, b, match_keys=match_keys) for a, b in zip(v1, v2)
        )
    else:
        return v1 == v2


class TestHtmlToContentState(TestCase):
    fixtures = ["test.json"]

    def assertContentStateEqual(self, v1, v2, match_keys=False):
        "Assert that two contentState structures are equal, ignoring 'key' properties if match_keys is False"
        self.assertTrue(
            content_state_equal(v1, v2, match_keys=match_keys),
            "%s does not match %s"
            % (json.dumps(v1, indent=4), json.dumps(v2, indent=4)),
        )

    def test_paragraphs(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p data-block-key='00000'>Hello world!</p>
            <p data-block-key='00001'>Goodbye world!</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Hello world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Goodbye world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00001",
                        "entityRanges": [],
                    },
                ],
            },
            match_keys=True,
        )

    def test_unknown_block_becomes_paragraph(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <foo>Hello world!</foo>
            <foo>I said hello world!</foo>
            <p>Goodbye world!</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Hello world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "I said hello world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Goodbye world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_bare_text_becomes_paragraph(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            before
            <p>paragraph</p>
            between
            <p>paragraph</p>
            after
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "before",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "paragraph",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "between",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "paragraph",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "after",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_ignore_unrecognised_tags_in_blocks(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p>Hello <foo>frabjuous</foo> world!</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Hello frabjuous world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_inline_styles(self):
        converter = ContentstateConverter(features=["bold", "italic"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>You <b>do <em>not</em> talk</b> about Fight Club.</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [
                            {"offset": 4, "length": 11, "style": "BOLD"},
                            {"offset": 7, "length": 3, "style": "ITALIC"},
                        ],
                        "text": "You do not talk about Fight Club.",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_inline_styles_at_top_level(self):
        converter = ContentstateConverter(features=["bold", "italic"])
        result = json.loads(
            converter.from_database_format(
                """
            You <b>do <em>not</em> talk</b> about Fight Club.
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [
                            {"offset": 4, "length": 11, "style": "BOLD"},
                            {"offset": 7, "length": 3, "style": "ITALIC"},
                        ],
                        "text": "You do not talk about Fight Club.",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_inline_styles_at_start_of_bare_block(self):
        converter = ContentstateConverter(features=["bold", "italic"])
        result = json.loads(
            converter.from_database_format(
                """<b>Seriously</b>, stop talking about <i>Fight Club</i> already."""
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [
                            {"offset": 0, "length": 9, "style": "BOLD"},
                            {"offset": 30, "length": 10, "style": "ITALIC"},
                        ],
                        "text": "Seriously, stop talking about Fight Club already.",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_inline_styles_depend_on_features(self):
        converter = ContentstateConverter(features=["italic", "just-made-it-up"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>You <b>do <em>not</em> talk</b> about Fight Club.</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [
                            {"offset": 7, "length": 3, "style": "ITALIC"}
                        ],
                        "text": "You do not talk about Fight Club.",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_ordered_list(self):
        converter = ContentstateConverter(features=["h1", "ol", "bold", "italic"])
        result = json.loads(
            converter.from_database_format(
                """
            <h1 data-block-key='00000'>The rules of Fight Club</h1>
            <ol>
                <li data-block-key='00001'>You do not talk about Fight Club.</li>
                <li data-block-key='00002'>You <b>do <em>not</em> talk</b> about Fight Club.</li>
            </ol>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "The rules of Fight Club",
                        "depth": 0,
                        "type": "header-one",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "You do not talk about Fight Club.",
                        "depth": 0,
                        "type": "ordered-list-item",
                        "key": "00001",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [
                            {"offset": 4, "length": 11, "style": "BOLD"},
                            {"offset": 7, "length": 3, "style": "ITALIC"},
                        ],
                        "text": "You do not talk about Fight Club.",
                        "depth": 0,
                        "type": "ordered-list-item",
                        "key": "00002",
                        "entityRanges": [],
                    },
                ],
            },
            match_keys=True,
        )

    def test_nested_list(self):
        converter = ContentstateConverter(features=["h1", "ul"])
        result = json.loads(
            converter.from_database_format(
                """
            <h1 data-block-key='00000'>Shopping list</h1>
            <ul>
                <li data-block-key='00001'>Milk</li>
                <li data-block-key='00002'>
                    Flour
                    <ul>
                        <li data-block-key='00003'>Plain</li>
                        <li data-block-key='00004'>Self-raising</li>
                    </ul>
                </li>
                <li data-block-key='00005'>Eggs</li>
            </ul>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Shopping list",
                        "depth": 0,
                        "type": "header-one",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Milk",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00001",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Flour",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00002",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Plain",
                        "depth": 1,
                        "type": "unordered-list-item",
                        "key": "00003",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Self-raising",
                        "depth": 1,
                        "type": "unordered-list-item",
                        "key": "00004",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Eggs",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00005",
                        "entityRanges": [],
                    },
                ],
            },
            match_keys=True,
        )

    def test_external_link(self):
        converter = ContentstateConverter(features=["link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>an <a href="http://wagtail.org">external</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "http://wagtail.org"},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an external link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_link_in_bare_text(self):
        converter = ContentstateConverter(features=["link"])
        result = json.loads(
            converter.from_database_format(
                """an <a href="http://wagtail.org">external</a> link"""
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "http://wagtail.org"},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an external link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_link_at_start_of_bare_text(self):
        converter = ContentstateConverter(features=["link"])
        result = json.loads(
            converter.from_database_format(
                """<a href="http://wagtail.org">an external link</a> and <a href="http://torchbox.com">another</a>"""
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "http://wagtail.org"},
                    },
                    "1": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "http://torchbox.com"},
                    },
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an external link and another",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [
                            {"offset": 0, "length": 16, "key": 0},
                            {"offset": 21, "length": 7, "key": 1},
                        ],
                    },
                ],
            },
        )

    def test_page_link(self):
        converter = ContentstateConverter(features=["link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>an <a linktype="page" id="3">internal</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"id": 3, "url": "/events/", "parentId": 2},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an internal link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_broken_page_link(self):
        converter = ContentstateConverter(features=["link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>an <a linktype="page" id="9999">internal</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {
                            "id": 9999,
                            "url": None,
                            "parentId": None,
                        },
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an internal link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_link_to_root_page(self):
        converter = ContentstateConverter(features=["link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>an <a linktype="page" id="1">internal</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"id": 1, "url": None, "parentId": None},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an internal link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_document_link(self):
        converter = ContentstateConverter(features=["document-link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>a <a linktype="document" id="1">document</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "DOCUMENT",
                        "data": {
                            "id": 1,
                            "url": "/documents/1/test.pdf",
                            "filename": "test.pdf",
                        },
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "a document link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 2, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_broken_document_link(self):
        converter = ContentstateConverter(features=["document-link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>a <a linktype="document" id="9999">document</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "DOCUMENT",
                        "data": {"id": 9999},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "a document link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 2, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_document_link_with_missing_id(self):
        converter = ContentstateConverter(features=["document-link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>a <a linktype="document">document</a> link</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {"mutability": "MUTABLE", "type": "DOCUMENT", "data": {}}
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "a document link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 2, "length": 8, "key": 0}],
                    },
                ],
            },
        )

    def test_image_embed(self):
        converter = ContentstateConverter(features=["image"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before</p>
            <embed embedtype="image" alt="an image" id="1" format="left" />
            <p>after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    }
                },
            },
        )

    def test_add_spacer_paragraph_between_image_embeds(self):
        converter = ContentstateConverter(features=["image"])
        result = json.loads(
            converter.from_database_format(
                """
            <embed embedtype="image" alt="an image" id="1" format="left" />
            <embed embedtype="image" alt="an image" id="1" format="left" />
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 1, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                    "1": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                },
            },
        )

    def test_image_after_list(self):
        """
        There should be no spacer paragraph inserted between a list and an image
        """
        converter = ContentstateConverter(features=["ul", "image"])
        result = json.loads(
            converter.from_database_format(
                """
            <ul>
                <li>Milk</li>
                <li>Eggs</li>
            </ul>
            <embed embedtype="image" alt="an image" id="1" format="left" />
            <ul>
                <li>More milk</li>
                <li>More eggs</li>
            </ul>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {
                    "0": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Milk",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Eggs",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 1, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "More milk",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "More eggs",
                        "depth": 0,
                        "type": "unordered-list-item",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    @patch("wagtail.embeds.embeds.get_embed")
    def test_media_embed(self, get_embed):
        get_embed.return_value = Embed(
            url="https://www.youtube.com/watch?v=Kh0Y2hVe_bw",
            max_width=None,
            type="video",
            html="test html",
            title="what are birds",
            author_name="look around you",
            provider_name="YouTube",
            thumbnail_url="http://test/thumbnail.url",
            width=1000,
            height=1000,
        )

        converter = ContentstateConverter(features=["embed"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before</p>
            <embed embedtype="media" url="https://www.youtube.com/watch?v=Kh0Y2hVe_bw" />
            <p>after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {
                            "thumbnail": "http://test/thumbnail.url",
                            "embedType": "video",
                            "providerName": "YouTube",
                            "title": "what are birds",
                            "authorName": "look around you",
                            "url": "https://www.youtube.com/watch?v=Kh0Y2hVe_bw",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "EMBED",
                    }
                },
            },
        )

    @patch("wagtail.embeds.embeds.get_embed")
    def test_add_spacer_paras_between_media_embeds(self, get_embed):
        get_embed.return_value = Embed(
            url="https://www.youtube.com/watch?v=Kh0Y2hVe_bw",
            max_width=None,
            type="video",
            html="test html",
            title="what are birds",
            author_name="look around you",
            provider_name="YouTube",
            thumbnail_url="http://test/thumbnail.url",
            width=1000,
            height=1000,
        )

        converter = ContentstateConverter(features=["embed"])
        result = json.loads(
            converter.from_database_format(
                """
            <embed embedtype="media" url="https://www.youtube.com/watch?v=Kh0Y2hVe_bw" />
            <embed embedtype="media" url="https://www.youtube.com/watch?v=Kh0Y2hVe_bw" />
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 1, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {
                            "thumbnail": "http://test/thumbnail.url",
                            "embedType": "video",
                            "providerName": "YouTube",
                            "title": "what are birds",
                            "authorName": "look around you",
                            "url": "https://www.youtube.com/watch?v=Kh0Y2hVe_bw",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "EMBED",
                    },
                    "1": {
                        "data": {
                            "thumbnail": "http://test/thumbnail.url",
                            "embedType": "video",
                            "providerName": "YouTube",
                            "title": "what are birds",
                            "authorName": "look around you",
                            "url": "https://www.youtube.com/watch?v=Kh0Y2hVe_bw",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "EMBED",
                    },
                },
            },
        )

    def test_hr(self):
        converter = ContentstateConverter(features=["hr"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before</p>
            <hr />
            <p>after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {},
                        "mutability": "IMMUTABLE",
                        "type": "HORIZONTAL_RULE",
                    }
                },
            },
        )

    def test_add_spacer_paragraph_between_hrs(self):
        converter = ContentstateConverter(features=["hr"])
        result = json.loads(
            converter.from_database_format(
                """
            <hr />
            <hr />
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 1, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {},
                        "mutability": "IMMUTABLE",
                        "type": "HORIZONTAL_RULE",
                    },
                    "1": {
                        "data": {},
                        "mutability": "IMMUTABLE",
                        "type": "HORIZONTAL_RULE",
                    },
                },
            },
        )

    def test_block_element_in_paragraph(self):
        converter = ContentstateConverter(features=["hr"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before<hr />after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {},
                        "mutability": "IMMUTABLE",
                        "type": "HORIZONTAL_RULE",
                    }
                },
            },
        )

    def test_br_element_in_paragraph(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before<br/>after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before\nafter",
                        "type": "unstyled",
                    }
                ],
            },
        )

    def test_br_element_between_paragraphs(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before</p>
            <br />
            <p>after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
            },
        )

    def test_block_element_in_empty_paragraph(self):
        converter = ContentstateConverter(features=["hr"])
        result = json.loads(
            converter.from_database_format(
                """
            <p><hr /></p>
            """
            )
        )
        # ignoring the paragraph completely would probably be better,
        # but we'll settle for an empty preceding paragraph and not crashing as the next best thing...
        # (and if it's the first/last block we actually do want a spacer paragraph anyhow)
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {},
                        "mutability": "IMMUTABLE",
                        "type": "HORIZONTAL_RULE",
                    }
                },
            },
        )

    def test_html_entities(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p>Arthur &quot;two sheds&quot; Jackson &lt;the third&gt; &amp; his wife</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": 'Arthur "two sheds" Jackson <the third> & his wife',
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_collapse_targeted_whitespace_characters(self):
        # We expect all targeted whitespace characters (one or more consecutively)
        # to be replaced by a single space. (\xa0 is a non-breaking whitespace)
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p>Multiple whitespaces:     should  be reduced</p>
            <p>Multiple non-breaking whitespace characters:  \xa0\xa0\xa0  should be preserved</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Multiple whitespaces: should be reduced",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Multiple non-breaking whitespace characters: \xa0\xa0\xa0 should be preserved",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_extra_end_tag_before(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            </p>
            <p>Before</p>
            """
            )
        )
        # The leading </p> tag should be ignored instead of blowing up with a
        # pop from empty list error
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Before",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_extra_end_tag_after(self):
        converter = ContentstateConverter(features=[])
        result = json.loads(
            converter.from_database_format(
                """
            <p>After</p>
            </p>
            """
            )
        )
        # The tailing </p> tag should be ignored instead of blowing up with a
        # pop from empty list error
        self.assertContentStateEqual(
            result,
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "After",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                ],
            },
        )

    def test_p_with_class(self):
        # Test support for custom conversion rules which require correct treatment of
        # CSS precedence in HTMLRuleset. Here, <p class="intro"> should match the
        # 'p[class="intro"]' rule rather than 'p' and thus become an 'intro-paragraph' block
        converter = ContentstateConverter(features=["intro"])
        result = json.loads(
            converter.from_database_format(
                """
            <p class="intro">before</p>
            <p>after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "intro-paragraph",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {},
            },
        )

    def test_image_inside_paragraph(self):
        # In Draftail's data model, images are block-level elements and therefore
        # split up preceding / following text into their own paragraphs
        converter = ContentstateConverter(features=["image"])
        result = json.loads(
            converter.from_database_format(
                """
            <p>before <embed embedtype="image" alt="an image" id="1" format="left" /> after</p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    }
                },
            },
        )

    def test_image_inside_style(self):
        # https://github.com/wagtail/wagtail/issues/4602 - ensure that an <embed> inside
        # an inline style is handled. This is not valid in Draftail as images are block-level,
        # but should be handled without errors, splitting the image into its own block
        converter = ContentstateConverter(features=["image", "italic"])
        result = json.loads(
            converter.from_database_format(
                """
            <p><i>before <embed embedtype="image" alt="an image" id="1" format="left" /> after</i></p>
            <p><i><embed embedtype="image" alt="an image" id="1" format="left" /></i></p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [
                            {"offset": 0, "length": 6, "style": "ITALIC"}
                        ],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [
                            {"offset": 0, "length": 5, "style": "ITALIC"}
                        ],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [
                            {"offset": 0, "length": 0, "style": "ITALIC"}
                        ],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 1, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [
                            {"offset": 0, "length": 0, "style": "ITALIC"}
                        ],
                        "entityRanges": [],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                    "1": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                },
            },
        )

    def test_image_inside_link(self):
        # https://github.com/wagtail/wagtail/issues/4602 - ensure that an <embed> inside
        # a link is handled. This is not valid in Draftail as images are block-level,
        # but should be handled without errors, splitting the image into its own block
        converter = ContentstateConverter(features=["image", "link"])
        result = json.loads(
            converter.from_database_format(
                """
            <p><a href="https://wagtail.org">before <embed embedtype="image" alt="an image" id="1" format="left" /> after</a></p>
            <p><a href="https://wagtail.org"><embed embedtype="image" alt="an image" id="1" format="left" /></a></p>
            """
            )
        )
        self.assertContentStateEqual(
            result,
            {
                "blocks": [
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 6}],
                        "depth": 0,
                        "text": "before",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 1, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 0, "offset": 0, "length": 5}],
                        "depth": 0,
                        "text": "after",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 2, "offset": 0, "length": 0}],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 3, "offset": 0, "length": 1}],
                        "depth": 0,
                        "text": " ",
                        "type": "atomic",
                    },
                    {
                        "key": "00000",
                        "inlineStyleRanges": [],
                        "entityRanges": [{"key": 2, "offset": 0, "length": 0}],
                        "depth": 0,
                        "text": "",
                        "type": "unstyled",
                    },
                ],
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "https://wagtail.org"},
                    },
                    "1": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                    "2": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "https://wagtail.org"},
                    },
                    "3": {
                        "data": {
                            "format": "left",
                            "alt": "an image",
                            "id": "1",
                            "src": "/media/not-found",
                        },
                        "mutability": "IMMUTABLE",
                        "type": "IMAGE",
                    },
                },
            },
        )


class TestContentStateToHtml(TestCase):
    def test_external_link(self):
        converter = ContentstateConverter(features=["link"])
        contentstate_json = json.dumps(
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "http://wagtail.org"},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an external link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            }
        )

        result = converter.to_database_format(contentstate_json)
        self.assertEqual(
            result,
            '<p data-block-key="00000">an <a href="http://wagtail.org">external</a> link</p>',
        )

    def test_local_link(self):
        converter = ContentstateConverter(features=["link"])
        contentstate_json = json.dumps(
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "/some/local/path/"},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an external link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            }
        )

        result = converter.to_database_format(contentstate_json)
        self.assertEqual(
            result,
            '<p data-block-key="00000">an <a href="/some/local/path/">external</a> link</p>',
        )

    def test_reject_javascript_link(self):
        converter = ContentstateConverter(features=["link"])
        contentstate_json = json.dumps(
            {
                "entityMap": {
                    "0": {
                        "mutability": "MUTABLE",
                        "type": "LINK",
                        "data": {"url": "javascript:alert('oh no')"},
                    }
                },
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "an external link",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [{"offset": 3, "length": 8, "key": 0}],
                    },
                ],
            }
        )

        result = converter.to_database_format(contentstate_json)
        self.assertEqual(
            result, '<p data-block-key="00000">an <a>external</a> link</p>'
        )

    def test_paragraphs_retain_keys(self):
        converter = ContentstateConverter(features=[])
        contentState = json.dumps(
            {
                "entityMap": {},
                "blocks": [
                    {
                        "inlineStyleRanges": [],
                        "text": "Hello world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00000",
                        "entityRanges": [],
                    },
                    {
                        "inlineStyleRanges": [],
                        "text": "Goodbye world!",
                        "depth": 0,
                        "type": "unstyled",
                        "key": "00001",
                        "entityRanges": [],
                    },
                ],
            }
        )
        result = converter.to_database_format(contentState)
        self.assertHTMLEqual(
            result,
            """
            <p data-block-key='00000'>Hello world!</p>
            <p data-block-key='00001'>Goodbye world!</p>
            """,
        )

    def test_wrapped_block_retains_key(self):
        # Test a block which uses a wrapper correctly receives the key defined on the inner element
        converter = ContentstateConverter(features=["h1", "ol", "bold", "italic"])
        result = converter.to_database_format(
            json.dumps(
                {
                    "entityMap": {},
                    "blocks": [
                        {
                            "inlineStyleRanges": [],
                            "text": "The rules of Fight Club",
                            "depth": 0,
                            "type": "header-one",
                            "key": "00000",
                            "entityRanges": [],
                        },
                        {
                            "inlineStyleRanges": [],
                            "text": "You do not talk about Fight Club.",
                            "depth": 0,
                            "type": "ordered-list-item",
                            "key": "00001",
                            "entityRanges": [],
                        },
                        {
                            "inlineStyleRanges": [],
                            "text": "You do not talk about Fight Club.",
                            "depth": 0,
                            "type": "ordered-list-item",
                            "key": "00002",
                            "entityRanges": [],
                        },
                    ],
                }
            )
        )
        self.assertHTMLEqual(
            result,
            """
            <h1 data-block-key='00000'>The rules of Fight Club</h1>
            <ol>
                <li data-block-key='00001'>You do not talk about Fight Club.</li>
                <li data-block-key='00002'>You do not talk about Fight Club.</li>
            </ol>
        """,
        )

    def test_wrap_block_function(self):
        # Draft JS exporter's block_map config can also contain a function to handle a particular block
        # Test that persist_key_for_block still works with such a function, making the resultant conversion
        # keep the same block key between html and contentstate
        exporter_config = {
            "block_map": {
                "unstyled": persist_key_for_block(
                    lambda props: DOM.create_element("p", {}, props["children"])
                ),
            },
            "style_map": {},
            "entity_decorators": {},
            "composite_decorators": [],
            "engine": DOM.STRING,
        }
        contentState = {
            "entityMap": {},
            "blocks": [
                {
                    "inlineStyleRanges": [],
                    "text": "Hello world!",
                    "depth": 0,
                    "type": "unstyled",
                    "key": "00000",
                    "entityRanges": [],
                },
                {
                    "inlineStyleRanges": [],
                    "text": "Goodbye world!",
                    "depth": 0,
                    "type": "unstyled",
                    "key": "00001",
                    "entityRanges": [],
                },
            ],
        }
        result = HTMLExporter(exporter_config).render(contentState)
        self.assertHTMLEqual(
            result,
            """
            <p data-block-key='00000'>Hello world!</p>
            <p data-block-key='00001'>Goodbye world!</p>
            """,
        )

    def test_style_fallback(self):
        # Test a block which uses an invalid inline style, and will be removed
        converter = ContentstateConverter(features=[])

        with self.assertLogs(level="WARNING") as log_output:
            result = converter.to_database_format(
                json.dumps(
                    {
                        "entityMap": {},
                        "blocks": [
                            {
                                "inlineStyleRanges": [
                                    {"offset": 0, "length": 12, "style": "UNDERLINE"}
                                ],
                                "text": "Hello world!",
                                "depth": 0,
                                "type": "unstyled",
                                "key": "00000",
                                "entityRanges": [],
                            },
                        ],
                    }
                )
            )

        self.assertHTMLEqual(
            result,
            """
            <p data-block-key="00000">
                Hello world!
            </p>
        """,
        )
        self.assertIn(
            'Missing config for "UNDERLINE". Deleting style.', log_output.output[0]
        )



from bs4 import BeautifulSoup
from django.test import TestCase

from wagtail.whitelist import (
    Whitelister,
    allow_without_attributes,
    attribute_rule,
    check_url,
)


class TestCheckUrl(TestCase):
    def test_allowed_url_schemes(self):
        for url_scheme in ["", "http", "https", "ftp", "mailto", "tel"]:
            url = url_scheme + "://www.example.com"
            self.assertTrue(bool(check_url(url)))

    def test_disallowed_url_scheme(self):
        self.assertFalse(bool(check_url("invalid://url")))

    def test_crafty_disallowed_url_scheme(self):
        """
        Some URL parsers do not parse 'jav\tascript:' as a valid scheme.
        Browsers, however, do. The checker needs to catch these crafty schemes
        """
        self.assertFalse(bool(check_url("jav\tascript:alert('XSS')")))


class TestAttributeRule(TestCase):
    def setUp(self):
        self.soup = BeautifulSoup('<b foo="bar">baz</b>', "html5lib")

    def test_no_rule_for_attr(self):
        """
        Test that attribute_rule() drops attributes for
        which no rule has been defined.
        """
        tag = self.soup.b
        fn = attribute_rule({"snowman": "barbecue"})
        fn(tag)
        self.assertEqual(str(tag), "<b>baz</b>")

    def test_rule_true_for_attr(self):
        """
        Test that attribute_rule() does not change attributes
        when the corresponding rule returns True
        """
        tag = self.soup.b
        fn = attribute_rule({"foo": True})
        fn(tag)
        self.assertEqual(str(tag), '<b foo="bar">baz</b>')

    def test_rule_false_for_attr(self):
        """
        Test that attribute_rule() drops attributes
        when the corresponding rule returns False
        """
        tag = self.soup.b
        fn = attribute_rule({"foo": False})
        fn(tag)
        self.assertEqual(str(tag), "<b>baz</b>")

    def test_callable_called_on_attr(self):
        """
        Test that when the rule returns a callable,
        attribute_rule() replaces the attribute with
        the result of calling the callable on the attribute.
        """
        tag = self.soup.b
        fn = attribute_rule({"foo": len})
        fn(tag)
        self.assertEqual(str(tag), '<b foo="3">baz</b>')

    def test_callable_returns_None(self):
        """
        Test that when the rule returns a callable,
        attribute_rule() replaces the attribute with
        the result of calling the callable on the attribute.
        """
        tag = self.soup.b
        fn = attribute_rule({"foo": lambda x: None})
        fn(tag)
        self.assertEqual(str(tag), "<b>baz</b>")

    def test_allow_without_attributes(self):
        """
        Test that attribute_rule() with will drop all
        attributes.
        """
        soup = BeautifulSoup(
            '<b foo="bar" baz="quux" snowman="barbecue"></b>', "html5lib"
        )
        tag = soup.b
        allow_without_attributes(tag)
        self.assertEqual(str(tag), "<b></b>")


class TestWhitelister(TestCase):
    def setUp(self):
        self.whitelister = Whitelister()

    def test_clean_unknown_node(self):
        """
        Unknown node should remove a node from the parent document
        """
        soup = BeautifulSoup("<foo><bar>baz</bar>quux</foo>", "html5lib")
        tag = soup.foo
        self.whitelister.clean_unknown_node("", soup.bar)
        self.assertEqual(str(tag), "<foo>quux</foo>")

    def test_clean_tag_node_cleans_nested_recognised_node(self):
        """
        <b> tags are allowed without attributes. This remains true
        when tags are nested.
        """
        soup = BeautifulSoup('<b><b class="delete me">foo</b></b>', "html5lib")
        tag = soup.b
        self.whitelister.clean_tag_node(tag, tag)
        self.assertEqual(str(tag), "<b><b>foo</b></b>")

    def test_clean_tag_node_disallows_nested_unrecognised_node(self):
        """
        <foo> tags should be removed, even when nested.
        """
        soup = BeautifulSoup("<b><foo>bar</foo></b>", "html5lib")
        tag = soup.b
        self.whitelister.clean_tag_node(tag, tag)
        self.assertEqual(str(tag), "<b>bar</b>")

    def test_clean_string_node_does_nothing(self):
        soup = BeautifulSoup("<b>bar</b>", "html5lib")
        string = soup.b.string
        self.whitelister.clean_string_node(string, string)
        self.assertEqual(str(string), "bar")

    def test_clean_node_does_not_change_navigable_strings(self):
        soup = BeautifulSoup("<b>bar</b>", "html5lib")
        string = soup.b.string
        self.whitelister.clean_node(string, string)
        self.assertEqual(str(string), "bar")

    def test_clean(self):
        """
        Whitelister.clean should remove disallowed tags and attributes from
        a string
        """
        string = '<b foo="bar">snowman <barbecue>Yorkshire</barbecue></b>'
        cleaned_string = self.whitelister.clean(string)
        self.assertEqual(cleaned_string, "<b>snowman Yorkshire</b>")

    def test_clean_comments(self):
        string = "<b>snowman Yorkshire<!--[if gte mso 10]>MS word junk<![endif]--></b>"
        cleaned_string = self.whitelister.clean(string)
        self.assertEqual(cleaned_string, "<b>snowman Yorkshire</b>")

    def test_quoting(self):
        string = '<img alt="Arthur &quot;two sheds&quot; Jackson" sheds="2">'
        cleaned_string = self.whitelister.clean(string)
        self.assertEqual(
            cleaned_string, '<img alt="Arthur &quot;two sheds&quot; Jackson"/>'
        )



import csv
import datetime
from collections import OrderedDict

from django.core.exceptions import FieldDoesNotExist
from django.http import HttpResponse, StreamingHttpResponse
from django.utils.dateformat import Formatter
from django.utils.encoding import force_str
from django.utils.formats import get_format
from django.utils.translation import gettext as _
from xlsxwriter.workbook import Workbook

from wagtail.admin.forms.search import SearchForm
from wagtail.coreutils import multigetattr
from wagtail.search.backends import get_search_backend
from wagtail.search.index import class_is_indexed


class SearchableListMixin:
    search_box_placeholder = _("Search")
    search_fields = None

    def get_search_form(self):
        return SearchForm(
            self.request.GET if self.request.GET.get("q") else None,
            placeholder=self.search_box_placeholder,
        )

    def get_queryset(self):
        queryset = super().get_queryset()
        search_form = self.get_search_form()

        if search_form.is_valid():
            q = search_form.cleaned_data["q"]

            if class_is_indexed(queryset.model):
                search_backend = get_search_backend()
                queryset = search_backend.search(q, queryset, fields=self.search_fields)
            else:
                filters = {
                    field + "__icontains": q for field in self.search_fields or []
                }

                queryset = queryset.filter(**filters)

        return queryset

    def get_context_data(self, **kwargs):
        if "search_form" not in kwargs:
            kwargs["search_form"] = self.get_search_form()
            kwargs["is_searching"] = bool(self.request.GET.get("q"))

        return super().get_context_data(**kwargs)


class Echo:
    """An object that implements just the write method of the file-like interface."""

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value.encode("UTF-8")


def list_to_str(value):
    return force_str(", ".join(value))


class ExcelDateFormatter(Formatter):
    data = None

    _formats = {
        "d": "DD",
        "j": "D",
        "D": "NN",
        "l": "NNNN",
        "S": "",
        "w": "",
        "z": "",
        "W": "",
        "m": "MM",
        "n": "M",
        "M": "MMM",
        "b": "MMM",
        "F": "MMMM",
        "E": "MMM",
        "N": "MMM.",
        "y": "YY",
        "Y": "YYYY",
        "L": "",
        "o": "",
        "g": "H",
        "G": "H",
        "h": "HH",
        "H": "HH",
        "i": "MM",
        "s": "SS",
        "u": "",
        "a": "AM/PM",
        "A": "AM/PM",
        "P": "HH:MM AM/PM",
        "e": "",
        "I": "",
        "O": "",
        "T": "",
        "Z": "",
        "c": "YYYY-MM-DD HH:MM:SS",
        "r": "NN, MMM D YY HH:MM:SS",
        "U": "[HH]:MM:SS",
    }

    def get(self):
        format = get_format("SHORT_DATETIME_FORMAT")
        return self.format(format)

    def __getattr__(self, name):
        if name in self._formats:
            return lambda: self._formats[name]
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )


class SpreadsheetExportMixin:
    """A mixin for views, providing spreadsheet export functionality in csv and xlsx formats"""

    FORMAT_XLSX = "xlsx"
    FORMAT_CSV = "csv"
    FORMATS = (FORMAT_XLSX, FORMAT_CSV)

    # A list of fields or callables (without arguments) to export from each item in the queryset (dotted paths allowed)
    list_export = []
    # A dictionary of custom preprocessing functions by field and format (expected value would be of the form {field_name: {format: function}})
    # If a valid field preprocessing function is found, any applicable value preprocessing functions will not be used
    custom_field_preprocess = {}
    # A dictionary of preprocessing functions by value class and format
    custom_value_preprocess = {
        (datetime.date, datetime.time): {FORMAT_XLSX: None},
        list: {FORMAT_CSV: list_to_str, FORMAT_XLSX: list_to_str},
    }
    # A dictionary of column heading overrides in the format {field: heading}
    export_headings = {}

    def get_filename(self):
        """Gets the base filename for the exported spreadsheet, without extensions"""
        return "spreadsheet-export"

    def to_row_dict(self, item):
        """Returns an OrderedDict (in the order given by list_export) of the exportable information for a model instance"""
        row_dict = OrderedDict(
            (field, multigetattr(item, field)) for field in self.list_export
        )
        return row_dict

    def get_preprocess_function(self, field, value, export_format):
        """Returns the preprocessing function for a given field name, field value, and export format"""

        # Try to find a field specific function and return it
        format_dict = self.custom_field_preprocess.get(field, {})
        if export_format in format_dict:
            return format_dict[export_format]

        # Otherwise check for a value class specific function
        for value_classes, format_dict in self.custom_value_preprocess.items():
            if isinstance(value, value_classes) and export_format in format_dict:
                return format_dict[export_format]

        # Finally resort to force_str to prevent encoding errors
        return force_str

    def write_xlsx_row(self, worksheet, row_dict, row_number):
        for col_number, (field, value) in enumerate(row_dict.items()):
            preprocess_function = self.get_preprocess_function(
                field, value, self.FORMAT_XLSX
            )
            processed_value = (
                preprocess_function(value) if preprocess_function else value
            )
            worksheet.write(row_number, col_number, processed_value)

    def write_csv_row(self, writer, row_dict):
        processed_row = {}
        for field, value in row_dict.items():
            preprocess_function = self.get_preprocess_function(
                field, value, self.FORMAT_CSV
            )
            processed_value = (
                preprocess_function(value) if preprocess_function else value
            )
            processed_row[field] = processed_value
        return writer.writerow(processed_row)

    def get_heading(self, queryset, field):
        """Get the heading label for a given field for a spreadsheet generated from queryset"""
        heading_override = self.export_headings.get(field)
        if heading_override:
            return force_str(heading_override)
        try:
            return force_str(queryset.model._meta.get_field(field).verbose_name.title())
        except (AttributeError, FieldDoesNotExist):
            return force_str(field)

    def stream_csv(self, queryset):
        """Generate a csv file line by line from queryset, to be used in a StreamingHTTPResponse"""
        writer = csv.DictWriter(Echo(), fieldnames=self.list_export)
        yield writer.writerow(
            {field: self.get_heading(queryset, field) for field in self.list_export}
        )

        for item in queryset:
            yield self.write_csv_row(writer, self.to_row_dict(item))

    def write_xlsx(self, queryset, output):
        """Write an xlsx workbook from a queryset"""
        workbook = Workbook(
            output,
            {
                "in_memory": True,
                "constant_memory": True,
                "remove_timezone": True,
                "default_date_format": ExcelDateFormatter().get(),
            },
        )
        worksheet = workbook.add_worksheet()

        for col_number, field in enumerate(self.list_export):
            worksheet.write(0, col_number, self.get_heading(queryset, field))

        for row_number, item in enumerate(queryset):
            self.write_xlsx_row(worksheet, self.to_row_dict(item), row_number + 1)

        workbook.close()

    def write_xlsx_response(self, queryset):
        response = HttpResponse(
            content_type="application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        )
        response["Content-Disposition"] = 'attachment; filename="{}.xlsx"'.format(
            self.get_filename()
        )
        self.write_xlsx(queryset, response)

        return response

    def write_csv_response(self, queryset):
        stream = self.stream_csv(queryset)

        response = StreamingHttpResponse(stream, content_type="text/csv")
        response["Content-Disposition"] = 'attachment; filename="{}.csv"'.format(
            self.get_filename()
        )
        return response

    def as_spreadsheet(self, queryset, spreadsheet_format):
        """Return a response with a spreadsheet representing the exported data from queryset, in the format specified"""
        if spreadsheet_format == self.FORMAT_CSV:
            return self.write_csv_response(queryset)
        elif spreadsheet_format == self.FORMAT_XLSX:
            return self.write_xlsx_response(queryset)

    def get_export_url(self, format):
        params = self.request.GET.copy()
        params["export"] = format
        return self.request.path + "?" + params.urlencode()

    @property
    def xlsx_export_url(self):
        return self.get_export_url("xlsx")

    @property
    def csv_export_url(self):
        return self.get_export_url("csv")



from django.urls import include, path

from wagtail import hooks
from wagtail.search.urls import admin as admin_urls


@hooks.register("register_admin_urls")
def register_admin_urls():
    return [
        path("search/", include(admin_urls, namespace="wagtailsearch_admin")),
    ]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class WagtailTestsAppConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    name = "wagtail.test.testapp"
    label = "tests"
    verbose_name = _("Wagtail tests")



from django.views.generic.base import ContextMixin, TemplateResponseMixin


class WagtailAdminTemplateMixin(TemplateResponseMixin, ContextMixin):
    """
    Mixin for views that render a template response using the standard Wagtail admin
    page furniture.
    Provides accessors for page title, subtitle and header icon.
    """

    page_title = ""
    page_subtitle = ""
    header_icon = ""
    template_name = "wagtailadmin/generic/base.html"

    def get_page_title(self):
        return self.page_title

    def get_page_subtitle(self):
        return self.page_subtitle

    def get_header_icon(self):
        return self.header_icon

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["page_title"] = self.get_page_title()
        context["page_subtitle"] = self.get_page_subtitle()
        context["header_icon"] = self.get_header_icon()
        return context



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0007_image_file_size"),
    ]

    operations = [
        migrations.AlterField(
            model_name="image",
            name="created_at",
            field=models.DateTimeField(
                db_index=True, verbose_name="Created at", auto_now_add=True
            ),
        ),
    ]



# Generated by Django 3.0.3 on 2020-03-03 15:26

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import modelcluster.fields


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
        ("wagtailcore", "0046_site_name_remove_null"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("auth", "0011_update_proxy_permissions"),
    ]

    operations = [
        migrations.CreateModel(
            name="Task",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, verbose_name="name")),
                (
                    "active",
                    models.BooleanField(
                        default=True,
                        help_text="Active tasks can be added to workflows. Deactivating a task does not remove it from existing workflows.",
                        verbose_name="active",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wagtail_tasks",
                        to="contenttypes.ContentType",
                        verbose_name="content type",
                    ),
                ),
            ],
            options={
                "verbose_name": "task",
                "verbose_name_plural": "tasks",
            },
        ),
        migrations.CreateModel(
            name="TaskState",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("in_progress", "In progress"),
                            ("approved", "Approved"),
                            ("rejected", "Rejected"),
                            ("skipped", "Skipped"),
                            ("cancelled", "Cancelled"),
                        ],
                        default="in_progress",
                        max_length=50,
                        verbose_name="status",
                    ),
                ),
                (
                    "started_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="started at"),
                ),
                (
                    "finished_at",
                    models.DateTimeField(
                        blank=True, null=True, verbose_name="finished at"
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="wagtail_task_states",
                        to="contenttypes.ContentType",
                        verbose_name="content type",
                    ),
                ),
                (
                    "page_revision",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="task_states",
                        to="wagtailcore.PageRevision",
                        verbose_name="page revision",
                    ),
                ),
                (
                    "task",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="task_states",
                        to="wagtailcore.Task",
                        verbose_name="task",
                    ),
                ),
            ],
            options={
                "verbose_name": "Task state",
                "verbose_name_plural": "Task states",
            },
        ),
        migrations.CreateModel(
            name="Workflow",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255, verbose_name="name")),
                (
                    "active",
                    models.BooleanField(
                        default=True,
                        help_text="Active workflows can be added to pages. Deactivating a workflow does not remove it from existing pages.",
                        verbose_name="active",
                    ),
                ),
            ],
            options={
                "verbose_name": "workflow",
                "verbose_name_plural": "workflows",
            },
        ),
        migrations.CreateModel(
            name="GroupApprovalTask",
            fields=[
                (
                    "task_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="wagtailcore.Task",
                    ),
                ),
            ],
            options={
                "verbose_name": "Group approval task",
                "verbose_name_plural": "Group approval tasks",
            },
            bases=("wagtailcore.task",),
        ),
        migrations.CreateModel(
            name="WorkflowState",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("in_progress", "In progress"),
                            ("approved", "Approved"),
                            ("rejected", "Rejected"),
                            ("cancelled", "Cancelled"),
                        ],
                        default="in_progress",
                        max_length=50,
                        verbose_name="status",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="created at"),
                ),
                (
                    "current_task_state",
                    models.OneToOneField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="wagtailcore.TaskState",
                        verbose_name="current task state",
                    ),
                ),
                (
                    "page",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="workflow_states",
                        to="wagtailcore.Page",
                        verbose_name="page",
                    ),
                ),
                (
                    "requested_by",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="requested_workflows",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="requested by",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="workflow_states",
                        to="wagtailcore.Workflow",
                        verbose_name="workflow",
                    ),
                ),
            ],
            options={
                "verbose_name": "Workflow state",
                "verbose_name_plural": "Workflow states",
            },
        ),
        migrations.CreateModel(
            name="WorkflowPage",
            fields=[
                (
                    "page",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        primary_key=True,
                        serialize=False,
                        to="wagtailcore.Page",
                        verbose_name="page",
                    ),
                ),
                (
                    "workflow",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="workflow_pages",
                        to="wagtailcore.Workflow",
                        verbose_name="workflow",
                    ),
                ),
            ],
            options={
                "verbose_name": "workflow page",
                "verbose_name_plural": "workflow pages",
            },
        ),
        migrations.AddField(
            model_name="taskstate",
            name="workflow_state",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="task_states",
                to="wagtailcore.WorkflowState",
                verbose_name="workflow state",
            ),
        ),
        migrations.CreateModel(
            name="WorkflowTask",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(blank=True, editable=False, null=True),
                ),
                (
                    "task",
                    models.ForeignKey(
                        limit_choices_to={"active": True},
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="workflow_tasks",
                        to="wagtailcore.Task",
                        verbose_name="task",
                    ),
                ),
                (
                    "workflow",
                    modelcluster.fields.ParentalKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="workflow_tasks",
                        to="wagtailcore.Workflow",
                        verbose_name="workflow_tasks",
                    ),
                ),
            ],
            options={
                "verbose_name": "workflow task order",
                "verbose_name_plural": "workflow task orders",
                "ordering": ["sort_order"],
                "abstract": False,
                "unique_together": {("workflow", "task")},
            },
        ),
        migrations.AddConstraint(
            model_name="workflowstate",
            constraint=models.UniqueConstraint(
                condition=models.Q(status="in_progress"),
                fields=("page",),
                name="unique_in_progress_workflow",
            ),
        ),
        migrations.AddField(
            model_name="groupapprovaltask",
            name="groups",
            field=models.ManyToManyField(
                help_text="Pages at this step in a workflow will be moderated or approved by these groups of users",
                to="auth.Group",
                verbose_name="groups",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailembeds", "0001_initial"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="embed",
            options={"verbose_name": "Embed"},
        ),
    ]



# Generated by Django 3.2.10 on 2022-01-19 10:48

from django.db import migrations, models
import django.db.models.deletion
import uuid
import wagtail.search.index


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0023_add_choose_permissions"),
        ("wagtailcore", "0066_collection_management_permissions"),
        ("modeladmintest", "0009_relatedlink"),
    ]

    operations = [
        migrations.CreateModel(
            name="TranslatableBook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "translation_key",
                    models.UUIDField(default=uuid.uuid4, editable=False),
                ),
                ("title", models.CharField(max_length=255)),
                (
                    "author",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="modeladmintest.author",
                    ),
                ),
                (
                    "cover_image",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="wagtailimages.image",
                    ),
                ),
                (
                    "locale",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="wagtailcore.locale",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "unique_together": {("translation_key", "locale")},
            },
            bases=(models.Model, wagtail.search.index.Indexed),
        ),
    ]



import json
from datetime import datetime
from urllib.parse import urljoin

from django import template
from django.conf import settings
from django.contrib.admin.utils import quote
from django.contrib.humanize.templatetags.humanize import intcomma, naturaltime
from django.contrib.messages.constants import DEFAULT_TAGS as MESSAGE_TAGS
from django.db.models import Min, QuerySet
from django.shortcuts import resolve_url as resolve_url_func
from django.template.defaultfilters import stringfilter
from django.templatetags.static import static
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils import timezone
from django.utils.encoding import force_str
from django.utils.html import avoid_wrapping, format_html, format_html_join, json_script
from django.utils.http import urlencode
from django.utils.safestring import mark_safe
from django.utils.timesince import timesince
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.localization import get_js_translation_strings
from wagtail.admin.menu import admin_menu
from wagtail.admin.navigation import get_explorable_root_page
from wagtail.admin.search import admin_search_areas
from wagtail.admin.staticfiles import versioned_static as versioned_static_func
from wagtail.admin.ui import sidebar
from wagtail.admin.utils import get_admin_base_url
from wagtail.admin.views.bulk_action.registry import bulk_action_registry
from wagtail.admin.views.pages.utils import get_valid_next_url_from_request
from wagtail.admin.widgets import ButtonWithDropdown, PageListingButton
from wagtail.coreutils import camelcase_to_underscore
from wagtail.coreutils import cautious_slugify as _cautious_slugify
from wagtail.coreutils import (
    escape_script,
    get_content_type_label,
    get_locales_display_names,
)
from wagtail.models import (
    Collection,
    CollectionViewRestriction,
    Locale,
    Page,
    PageViewRestriction,
    UserPagePermissionsProxy,
)
from wagtail.telepath import JSContext
from wagtail.users.utils import get_gravatar_url

register = template.Library()

register.filter("intcomma", intcomma)
register.filter("naturaltime", naturaltime)


@register.inclusion_tag("wagtailadmin/shared/breadcrumb.html", takes_context=True)
def explorer_breadcrumb(
    context,
    page,
    page_perms=None,
    include_self=True,
    use_next_template=False,
    trailing_breadcrumb_title=None,
):
    user = context["request"].user

    # find the closest common ancestor of the pages that this user has direct explore permission
    # (i.e. add/edit/publish/lock) over; this will be the root of the breadcrumb
    cca = get_explorable_root_page(user)
    if not cca:
        return {"pages": Page.objects.none()}

    return {
        "pages": page.get_ancestors(inclusive=include_self)
        .descendant_of(cca, inclusive=True)
        .specific(),
        "current_page": page,
        "page_perms": page_perms,
        "use_next_template": use_next_template,
        "trailing_breadcrumb_title": trailing_breadcrumb_title,  # Only used in collapsible breadcrumb templates
    }


@register.inclusion_tag("wagtailadmin/shared/move_breadcrumb.html", takes_context=True)
def move_breadcrumb(context, page_to_move, viewed_page):
    user = context["request"].user
    cca = get_explorable_root_page(user)
    if not cca:
        return {"pages": Page.objects.none()}

    return {
        "pages": viewed_page.get_ancestors(inclusive=True)
        .descendant_of(cca, inclusive=True)
        .specific(),
        "page_to_move_id": page_to_move.id,
    }


@register.inclusion_tag("wagtailadmin/shared/search_other.html", takes_context=True)
def search_other(context, current=None):
    request = context["request"]

    return {
        "options_html": admin_search_areas.render_html(request, current),
        "request": request,
    }


@register.filter("ellipsistrim")
def ellipsistrim(value, max_length):
    if len(value) > max_length:
        truncd_val = value[:max_length]
        if not len(value) == (max_length + 1) and value[max_length + 1] != " ":
            truncd_val = truncd_val[: truncd_val.rfind(" ")]
        return truncd_val + "…"
    return value


@register.filter
def fieldtype(bound_field):
    try:
        return camelcase_to_underscore(bound_field.field.__class__.__name__)
    except AttributeError:
        try:
            return camelcase_to_underscore(bound_field.__class__.__name__)
        except AttributeError:
            return ""


@register.filter
def widgettype(bound_field):
    try:
        return camelcase_to_underscore(bound_field.field.widget.__class__.__name__)
    except AttributeError:
        try:
            return camelcase_to_underscore(bound_field.widget.__class__.__name__)
        except AttributeError:
            return ""


def _get_user_page_permissions(context):
    # Create a UserPagePermissionsProxy object to represent the user's global permissions, and
    # cache it in the context for the duration of the page request, if one does not exist already
    if "user_page_permissions" not in context:
        context["user_page_permissions"] = UserPagePermissionsProxy(
            context["request"].user
        )

    return context["user_page_permissions"]


@register.simple_tag(takes_context=True)
def page_permissions(context, page):
    """
    Usage: {% page_permissions page as page_perms %}
    Sets the variable 'page_perms' to a PagePermissionTester object that can be queried to find out
    what actions the current logged-in user can perform on the given page.
    """
    return _get_user_page_permissions(context).for_page(page)


@register.simple_tag(takes_context=True)
def test_collection_is_public(context, collection):
    """
    Usage: {% test_collection_is_public collection as is_public %}
    Sets 'is_public' to True iff there are no collection view restrictions in place
    on this collection.
    Caches the list of collection view restrictions in the context, to avoid repeated
    DB queries on repeated calls.
    """
    if "all_collection_view_restrictions" not in context:
        context[
            "all_collection_view_restrictions"
        ] = CollectionViewRestriction.objects.select_related("collection").values_list(
            "collection__name", flat=True
        )

    is_private = collection.name in context["all_collection_view_restrictions"]

    return not is_private


@register.simple_tag(takes_context=True)
def test_page_is_public(context, page):
    """
    Usage: {% test_page_is_public page as is_public %}
    Sets 'is_public' to True iff there are no page view restrictions in place on
    this page.
    Caches the list of page view restrictions on the request, to avoid repeated
    DB queries on repeated calls.
    """
    if not hasattr(context["request"], "all_page_view_restriction_paths"):
        context[
            "request"
        ].all_page_view_restriction_paths = PageViewRestriction.objects.select_related(
            "page"
        ).values_list(
            "page__path", flat=True
        )

    is_private = any(
        [
            page.path.startswith(restricted_path)
            for restricted_path in context["request"].all_page_view_restriction_paths
        ]
    )

    return not is_private


@register.simple_tag
def hook_output(hook_name):
    """
    Example: {% hook_output 'insert_editor_css' %}
    Whenever we have a hook whose functions take no parameters and return a string, this tag can be used
    to output the concatenation of all of those return values onto the page.
    Note that the output is not escaped - it is the hook function's responsibility to escape unsafe content.
    """
    snippets = [fn() for fn in hooks.get_hooks(hook_name)]
    return mark_safe("".join(snippets))


@register.simple_tag
def usage_count_enabled():
    return getattr(settings, "WAGTAIL_USAGE_COUNT_ENABLED", False)


@register.simple_tag
def base_url_setting(default=None):
    return get_admin_base_url() or default


@register.simple_tag
def allow_unicode_slugs():
    return getattr(settings, "WAGTAIL_ALLOW_UNICODE_SLUGS", True)


@register.simple_tag
def auto_update_preview():
    return getattr(settings, "WAGTAIL_AUTO_UPDATE_PREVIEW", False)


class EscapeScriptNode(template.Node):
    TAG_NAME = "escapescript"

    def __init__(self, nodelist):
        super().__init__()
        self.nodelist = nodelist

    def render(self, context):
        out = self.nodelist.render(context)
        return escape_script(out)

    @classmethod
    def handle(cls, parser, token):
        nodelist = parser.parse(("end" + EscapeScriptNode.TAG_NAME,))
        parser.delete_first_token()
        return cls(nodelist)


register.tag(EscapeScriptNode.TAG_NAME, EscapeScriptNode.handle)


# Helpers for Widget.render_with_errors, our extension to the Django widget API that allows widgets to
# take on the responsibility of rendering their own error messages
@register.filter
def render_with_errors(bound_field):
    """
    Usage: {{ field|render_with_errors }} as opposed to {{ field }}.
    If the field (a BoundField instance) has errors on it, and the associated widget implements
    a render_with_errors method, call that; otherwise, call the regular widget rendering mechanism.
    """
    widget = bound_field.field.widget
    if bound_field.errors and hasattr(widget, "render_with_errors"):
        return widget.render_with_errors(
            bound_field.html_name,
            bound_field.value(),
            attrs={"id": bound_field.auto_id},
            errors=bound_field.errors,
        )
    else:
        return bound_field.as_widget()


@register.filter
def has_unrendered_errors(bound_field):
    """
    Return true if this field has errors that were not accounted for by render_with_errors, because
    the widget does not support the render_with_errors method
    """
    return bound_field.errors and not hasattr(
        bound_field.field.widget, "render_with_errors"
    )


@register.filter(is_safe=True)
@stringfilter
def cautious_slugify(value):
    return _cautious_slugify(value)


@register.simple_tag(takes_context=True)
def querystring(context, **kwargs):
    """
    Print out the current querystring. Any keyword arguments to this template
    tag will be added to the querystring before it is printed out.

        <a href="/page/{% querystring key='value' %}">

    Will result in something like:

        <a href="/page/?foo=bar&key=value">
    """
    request = context["request"]
    querydict = request.GET.copy()
    # Can't do querydict.update(kwargs), because QueryDict.update() appends to
    # the list of values, instead of replacing the values.
    for key, value in kwargs.items():
        if value is None:
            # Remove the key if the value is None
            querydict.pop(key, None)
        else:
            # Set the key otherwise
            querydict[key] = str(value)

    return "?" + querydict.urlencode()


@register.simple_tag(takes_context=True)
def page_table_header_label(context, label=None, parent_page_title=None, **kwargs):
    """
    Wraps table_header_label to add a title attribute based on the parent page title and the column label
    """
    if label:
        translation_context = {"parent": parent_page_title, "label": label}
        ascending_title_text = (
            _(
                "Sort the order of child pages within '%(parent)s' by '%(label)s' in ascending order."
            )
            % translation_context
        )
        descending_title_text = (
            _(
                "Sort the order of child pages within '%(parent)s' by '%(label)s' in descending order."
            )
            % translation_context
        )
    else:
        ascending_title_text = None
        descending_title_text = None

    return table_header_label(
        context,
        label=label,
        ascending_title_text=ascending_title_text,
        descending_title_text=descending_title_text,
        **kwargs,
    )


@register.simple_tag(takes_context=True)
def table_header_label(
    context,
    label=None,
    sortable=True,
    ordering=None,
    sort_context_var="ordering",
    sort_param="ordering",
    sort_field=None,
    ascending_title_text=None,
    descending_title_text=None,
):
    """
    A label to go in a table header cell, optionally with a 'sort' link that alternates between
    forward and reverse sorting

    label = label text
    ordering = current active ordering. If not specified, we will fetch it from the template context variable
        given by sort_context_var. (We don't fetch it from the URL because that wouldn't give the view method
        the opportunity to set a default)
    sort_param = URL parameter that indicates the current active ordering
    sort_field = the value for sort_param that indicates that sorting is currently on this column.
        For example, if sort_param='ordering' and sort_field='title', then a URL parameter of
        ordering=title indicates that the listing is ordered forwards on this column, and a URL parameter
        of ordering=-title indicated that the listing is ordered in reverse on this column
    ascending_title_text = title attribute to use on the link when the link action will sort in ascending order
    descending_title_text = title attribute to use on the link when the link action will sort in descending order

    To disable sorting on this column, set sortable=False or leave sort_field unspecified.
    """
    if not sortable or not sort_field:
        # render label without a sort link
        return label

    if ordering is None:
        ordering = context.get(sort_context_var)
    reverse_sort_field = "-%s" % sort_field

    if ordering == sort_field:
        # currently ordering forwards on this column; link should change to reverse ordering
        attrs = {
            "href": querystring(context, **{sort_param: reverse_sort_field}),
            "class": "icon icon-arrow-down-after teal",
        }
        if descending_title_text is not None:
            attrs["title"] = descending_title_text

    elif ordering == reverse_sort_field:
        # currently ordering backwards on this column; link should change to forward ordering
        attrs = {
            "href": querystring(context, **{sort_param: sort_field}),
            "class": "icon icon-arrow-up-after teal",
        }
        if ascending_title_text is not None:
            attrs["title"] = ascending_title_text

    else:
        # not currently ordering on this column; link should change to forward ordering
        attrs = {
            "href": querystring(context, **{sort_param: sort_field}),
            "class": "icon icon-arrow-down-after",
        }
        if ascending_title_text is not None:
            attrs["title"] = ascending_title_text

    attrs_string = format_html_join(" ", '{}="{}"', attrs.items())

    return format_html(
        # need whitespace around label for correct positioning of arrow icon
        "<a {attrs}> {label} </a>",
        attrs=attrs_string,
        label=label,
    )


@register.simple_tag(takes_context=True)
def pagination_querystring(context, page_number, page_key="p"):
    """
    Print out a querystring with an updated page number:

        {% if page.has_next_page %}
            <a href="{% pagination_link page.next_page_number %}">Next page</a>
        {% endif %}
    """
    return querystring(context, **{page_key: page_number})


@register.inclusion_tag(
    "wagtailadmin/pages/listing/_pagination.html", takes_context=True
)
def paginate(context, page, base_url="", page_key="p", classnames=""):
    """
    Print pagination previous/next links, and the page count. Take the
    following arguments:

    page
        The current page of results. This should be a Django pagination `Page`
        instance

    base_url
        The base URL of the next/previous page, with no querystring.
        This is optional, and defaults to the current page by just printing the
        querystring for the next/previous page.

    page_key
        The name of the page variable in the query string. Defaults to 'p'.

    classnames
        Extra classes to add to the next/previous links.
    """
    request = context["request"]
    return {
        "base_url": base_url,
        "classnames": classnames,
        "request": request,
        "page": page,
        "page_key": page_key,
        "paginator": page.paginator,
    }


@register.inclusion_tag("wagtailadmin/pages/listing/_buttons.html", takes_context=True)
def page_listing_buttons(context, page, page_perms, is_parent=False):
    next_url = context.request.path
    button_hooks = hooks.get_hooks("register_page_listing_buttons")

    buttons = []
    for hook in button_hooks:
        buttons.extend(hook(page, page_perms, is_parent, next_url))

    buttons.sort()

    for hook in hooks.get_hooks("construct_page_listing_buttons"):
        hook(buttons, page, page_perms, is_parent, context)

    return {"page": page, "buttons": buttons}


@register.inclusion_tag(
    "wagtailadmin/pages/listing/_modern_dropdown.html", takes_context=True
)
def page_header_buttons(context, page, page_perms):
    next_url = context.request.path
    button_hooks = hooks.get_hooks("register_page_header_buttons")

    buttons = []
    for hook in button_hooks:
        buttons.extend(hook(page, page_perms, next_url))

    buttons.sort()
    return {
        "page": page,
        "buttons": buttons,
        "title": _("Actions"),
        "icon_name": "dots-horizontal",
        "classes": [
            "w-flex",
            "w-justify-center",
            "w-items-center",
            "w-h-[50px]",
        ],
        "button_classes": [
            "w-p-0",
            "w-w-12",
            "w-h-full",
            "w-text-primary",
            "w-bg-transparent",
            "hover:w-scale-110",
            "w-transition",
            "w-outline-offset-inside",
            "w-relative",
            "w-z-30",
        ],
        "hide_title": True,
    }


@register.inclusion_tag("wagtailadmin/pages/listing/_buttons.html", takes_context=True)
def bulk_action_choices(context, app_label, model_name):

    bulk_actions_list = list(
        bulk_action_registry.get_bulk_actions_for_model(app_label, model_name)
    )
    bulk_actions_list.sort(key=lambda x: x.action_priority)

    bulk_action_more_list = []
    if len(bulk_actions_list) > 4:
        bulk_action_more_list = bulk_actions_list[4:]
        bulk_actions_list = bulk_actions_list[:4]

    next_url = get_valid_next_url_from_request(context["request"])
    if not next_url:
        next_url = context["request"].path

    bulk_action_buttons = [
        PageListingButton(
            action.display_name,
            reverse(
                "wagtail_bulk_action", args=[app_label, model_name, action.action_type]
            )
            + "?"
            + urlencode({"next": next_url}),
            attrs={"aria-label": action.aria_label},
            priority=action.action_priority,
            classes=action.classes | {"bulk-action-btn"},
        )
        for action in bulk_actions_list
    ]

    if bulk_action_more_list:
        more_button = ButtonWithDropdown(
            label=_("More"),
            attrs={"title": _("View more bulk actions")},
            classes={"bulk-actions-more", "dropup"},
            button_classes={"button", "button-small"},
            buttons_data=[
                {
                    "label": action.display_name,
                    "url": reverse(
                        "wagtail_bulk_action",
                        args=[app_label, model_name, action.action_type],
                    )
                    + "?"
                    + urlencode({"next": next_url}),
                    "attrs": {"aria-label": action.aria_label},
                    "priority": action.action_priority,
                    "classes": {"bulk-action-btn"},
                }
                for action in bulk_action_more_list
            ],
        )
        more_button.is_parent = True
        bulk_action_buttons.append(more_button)

    return {"buttons": bulk_action_buttons}


@register.simple_tag
def message_level_tag(message):
    """
    Return the tag for this message's level as defined in
    django.contrib.messages.constants.DEFAULT_TAGS, ignoring the project-level
    MESSAGE_TAGS setting (which end-users might customise).
    """
    return MESSAGE_TAGS.get(message.level)


@register.simple_tag
def message_tags(message):
    level_tag = message_level_tag(message)
    if message.extra_tags and level_tag:
        return message.extra_tags + " " + level_tag
    elif message.extra_tags:
        return message.extra_tags
    elif level_tag:
        return level_tag
    else:
        return ""


@register.filter("abs")
def _abs(val):
    return abs(val)


@register.filter
def admin_urlquote(value):
    return quote(value)


@register.simple_tag
def avatar_url(user, size=50, gravatar_only=False):
    """
    A template tag that receives a user and size and return
    the appropriate avatar url for that user.
    Example usage: {% avatar_url request.user 50 %}
    """

    if (
        not gravatar_only
        and hasattr(user, "wagtail_userprofile")
        and user.wagtail_userprofile.avatar
    ):
        return user.wagtail_userprofile.avatar.url

    if hasattr(user, "email"):
        gravatar_url = get_gravatar_url(user.email, size=size)
        if gravatar_url is not None:
            return gravatar_url

    return versioned_static_func("wagtailadmin/images/default-user-avatar.png")


@register.simple_tag
def js_translation_strings():
    return mark_safe(json.dumps(get_js_translation_strings()))


@register.simple_tag
def notification_static(path):
    """
    Variant of the {% static %}` tag for use in notification emails - tries to form
    a full URL using WAGTAILADMIN_BASE_URL if the static URL isn't already a full URL.
    """
    return urljoin(base_url_setting(), static(path))


@register.simple_tag
def versioned_static(path):
    """
    Wrapper for Django's static file finder to append a cache-busting query parameter
    that updates on each Wagtail version
    """
    return versioned_static_func(path)


@register.inclusion_tag("wagtailadmin/shared/icon.html", takes_context=False)
def icon(name=None, class_name="icon", title=None, wrapped=False):
    """
    Abstracts away the actual icon implementation.

    Usage:
        {% load wagtailadmin_tags %}
        ...
        {% icon name="cogs" class_name="icon--red" title="Settings" %}

    :param name: the icon name/id, required (string)
    :param class_name: default 'icon' (string)
    :param title: accessible label intended for screen readers (string)
    :return: Rendered template snippet (string)
    """
    if not name:
        raise ValueError("You must supply an icon name")

    return {"name": name, "class_name": class_name, "title": title, "wrapped": wrapped}


@register.filter()
def timesince_simple(d):
    """
    Returns a simplified timesince:
    19 hours, 48 minutes ago -> 19 hours ago
    1 week, 1 day ago -> 1 week ago
    0 minutes ago -> just now
    """
    time_period = timesince(d).split(",")[0]
    if time_period == avoid_wrapping(_("0 minutes")):
        return _("Just now")
    return _("%(time_period)s ago") % {"time_period": time_period}


@register.simple_tag
def timesince_last_update(
    last_update, time_prefix="", user_display_name="", use_shorthand=True
):
    """
    Returns:
         - the time of update if last_update is today, if any prefix is supplied, the output will use it
         - time since last update otherwise. Defaults to the simplified timesince,
           but can return the full string if needed
    """
    if last_update.date() == datetime.today().date():
        if timezone.is_aware(last_update):
            time_str = timezone.localtime(last_update).strftime("%H:%M")
        else:
            time_str = last_update.strftime("%H:%M")

        time_prefix = f"{time_prefix} " if time_prefix else time_prefix
        by_user = f" by {user_display_name}" if user_display_name else user_display_name

        return f"{time_prefix}{time_str}{by_user}"

    else:
        if use_shorthand:
            return timesince_simple(last_update)
        return _("%(time_period)s ago") % {"time_period": timesince(last_update)}


@register.simple_tag
def format_collection(coll: Collection, min_depth: int = 2) -> str:
    """
    Renders a given Collection's name as a formatted string that displays its
    hierarchical depth via indentation. If min_depth is supplied, the
    Collection's depth is rendered relative to that depth. min_depth defaults
    to 2, the depth of the first non-Root Collection.

    Example usage: {% format_collection collection min_depth %}
    Example output: "&nbsp;&nbsp;&nbsp;&nbsp;&#x21b3 Child Collection"
    """
    return coll.get_indented_name(min_depth, html=True)


@register.simple_tag
def minimum_collection_depth(collections: QuerySet) -> int:
    """
    Returns the minimum depth of the Collections in the given queryset.
    Call this before beginning a loop through Collections that will
    use {% format_collection collection min_depth %}.
    """
    return collections.aggregate(Min("depth"))["depth__min"] or 2


@register.filter
def user_display_name(user):
    """
    Returns the preferred display name for the given user object: the result of
    user.get_full_name() if implemented and non-empty, or user.get_username() otherwise.
    """
    try:
        full_name = user.get_full_name().strip()
        if full_name:
            return full_name
    except AttributeError:
        pass

    try:
        return user.get_username()
    except AttributeError:
        # we were passed None or something else that isn't a valid user object; return
        # empty string to replicate the behaviour of {{ user.get_full_name|default:user.get_username }}
        return ""


@register.filter
def format_content_type(content_type):
    return get_content_type_label(content_type)


@register.simple_tag
def i18n_enabled():
    return getattr(settings, "WAGTAIL_I18N_ENABLED", False)


@register.simple_tag
def locales():
    return json.dumps(
        [
            {
                "code": locale.language_code,
                "display_name": force_str(locale.get_display_name()),
            }
            for locale in Locale.objects.all()
        ]
    )


@register.simple_tag
def locale_label_from_id(locale_id):
    """
    Returns the Locale display name given its id.
    """
    return get_locales_display_names().get(locale_id)


@register.simple_tag(takes_context=True)
def sidebar_collapsed(context):
    request = context.get("request")
    collapsed = request.COOKIES.get("wagtail_sidebar_collapsed", "0")
    if collapsed == "0":
        return False
    return True


@register.simple_tag(takes_context=True)
def sidebar_props(context):
    request = context["request"]
    search_areas = admin_search_areas.search_items_for_request(request)
    if search_areas:
        search_area = search_areas[0]
    else:
        search_area = None

    account_menu = [
        sidebar.LinkMenuItem(
            "account", _("Account"), reverse("wagtailadmin_account"), icon_name="user"
        ),
        sidebar.LinkMenuItem(
            "logout", _("Log out"), reverse("wagtailadmin_logout"), icon_name="logout"
        ),
    ]

    modules = [
        sidebar.WagtailBrandingModule(),
        sidebar.SearchModule(search_area) if search_area else None,
        sidebar.MainMenuModule(
            admin_menu.render_component(request), account_menu, request.user
        ),
    ]
    modules = [module for module in modules if module is not None]

    return json_script(
        {
            "modules": JSContext().pack(modules),
        },
        element_id="wagtail-sidebar-props",
    )


@register.simple_tag
def get_comments_enabled():
    return getattr(settings, "WAGTAILADMIN_COMMENTS_ENABLED", True)


@register.simple_tag
def resolve_url(url):
    # Used by wagtailadmin/shared/pagination_nav.html - given an input that may be a URL route
    # name, or a direct URL path, return it as a direct URL path. On failure (or being passed
    # an empty / None value), return empty string
    if not url:
        return ""

    try:
        return resolve_url_func(url)
    except NoReverseMatch:
        return ""


@register.simple_tag(takes_context=True)
def component(context, obj, fallback_render_method=False):
    # Render a component by calling its render_html method, passing request and context from the
    # calling template.
    # If fallback_render_method is true, objects without a render_html method will have render()
    # called instead (with no arguments) - this is to provide deprecation path for things that have
    # been newly upgraded to use the component pattern.

    has_render_html_method = hasattr(obj, "render_html")
    if fallback_render_method and not has_render_html_method and hasattr(obj, "render"):
        return obj.render()
    elif not has_render_html_method:
        raise ValueError("Cannot render %r as a component" % (obj,))

    return obj.render_html(context)



import warnings
from operator import itemgetter

import l18n
from django import forms
from django.contrib.auth import get_user_model
from django.db.models.fields import BLANK_CHOICE_DASH
from django.utils.translation import get_language_info
from django.utils.translation import gettext_lazy as _

from wagtail.admin.localization import (
    get_available_admin_languages,
    get_available_admin_time_zones,
)
from wagtail.admin.widgets import SwitchInput
from wagtail.models import UserPagePermissionsProxy
from wagtail.users.models import UserProfile

User = get_user_model()


class NotificationPreferencesForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        user_perms = UserPagePermissionsProxy(self.instance.user)
        if not user_perms.can_publish_pages():
            del self.fields["submitted_notifications"]
        if not user_perms.can_edit_pages():
            del self.fields["approved_notifications"]
            del self.fields["rejected_notifications"]
            del self.fields["updated_comments_notifications"]

    class Meta:
        model = UserProfile
        fields = [
            "submitted_notifications",
            "approved_notifications",
            "rejected_notifications",
            "updated_comments_notifications",
        ]
        widgets = {
            "submitted_notifications": SwitchInput(),
            "approved_notifications": SwitchInput(),
            "rejected_notifications": SwitchInput(),
            "updated_comments_notifications": SwitchInput(),
        }


def _get_language_choices():
    language_choices = [
        (lang_code, get_language_info(lang_code)["name_local"])
        for lang_code, lang_name in get_available_admin_languages()
    ]
    return sorted(BLANK_CHOICE_DASH + language_choices, key=lambda l: l[1].lower())


def _get_time_zone_choices():
    time_zones = [
        (tz, str(l18n.tz_fullnames.get(tz, tz)))
        for tz in get_available_admin_time_zones()
    ]
    time_zones.sort(key=itemgetter(1))
    return BLANK_CHOICE_DASH + time_zones


class LocalePreferencesForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if len(get_available_admin_languages()) <= 1:
            del self.fields["preferred_language"]

        if len(get_available_admin_time_zones()) <= 1:
            del self.fields["current_time_zone"]

    preferred_language = forms.ChoiceField(
        required=False, choices=_get_language_choices, label=_("Preferred language")
    )

    current_time_zone = forms.ChoiceField(
        required=False, choices=_get_time_zone_choices, label=_("Current time zone")
    )

    class Meta:
        model = UserProfile
        fields = ["preferred_language", "current_time_zone"]


class NameEmailForm(forms.ModelForm):
    first_name = forms.CharField(required=True, label=_("First Name"))
    last_name = forms.CharField(required=True, label=_("Last Name"))
    email = forms.EmailField(required=True, label=_("Email"))

    def __init__(self, *args, **kwargs):
        from wagtail.admin.views.account import email_management_enabled

        super().__init__(*args, **kwargs)

        if not email_management_enabled():
            del self.fields["email"]

    class Meta:
        model = User
        fields = ["first_name", "last_name", "email"]


class AvatarPreferencesForm(forms.ModelForm):
    avatar = forms.ImageField(label=_("Upload a profile picture"), required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._original_avatar = self.instance.avatar

    def save(self, commit=True):
        if (
            commit
            and self._original_avatar
            and (self._original_avatar != self.cleaned_data["avatar"])
        ):
            # Call delete() on the storage backend directly, as calling self._original_avatar.delete()
            # will clear the now-updated field on self.instance too
            try:
                self._original_avatar.storage.delete(self._original_avatar.name)
            except IOError:
                # failure to delete the old avatar shouldn't prevent us from continuing
                warnings.warn(
                    "Failed to delete old avatar file: %s" % self._original_avatar.name
                )
        super().save(commit=commit)

    class Meta:
        model = UserProfile
        fields = ["avatar"]



from django.core.exceptions import PermissionDenied
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.decorators.http import require_GET

from wagtail.admin import messages
from wagtail.admin.mail import send_moderation_notification
from wagtail.models import PageRevision


def approve_moderation(request, revision_id):
    revision = get_object_or_404(PageRevision, id=revision_id)
    if not revision.page.permissions_for_user(request.user).can_publish():
        raise PermissionDenied

    if not revision.submitted_for_moderation:
        messages.error(
            request,
            _("The page '{0}' is not currently awaiting moderation.").format(
                revision.page.specific_deferred.get_admin_display_title()
            ),
        )
        return redirect("wagtailadmin_home")

    if request.method == "POST":
        revision.approve_moderation(user=request.user)

        message = _("Page '{0}' published.").format(
            revision.page.specific_deferred.get_admin_display_title()
        )
        buttons = []
        if revision.page.url is not None:
            buttons.append(
                messages.button(revision.page.url, _("View live"), new_window=False)
            )
        buttons.append(
            messages.button(
                reverse("wagtailadmin_pages:edit", args=(revision.page.id,)), _("Edit")
            )
        )
        messages.success(request, message, buttons=buttons)

        if not send_moderation_notification(revision, "approved", request.user):
            messages.error(request, _("Failed to send approval notifications"))

    return redirect("wagtailadmin_home")


def reject_moderation(request, revision_id):
    revision = get_object_or_404(PageRevision, id=revision_id)
    if not revision.page.permissions_for_user(request.user).can_publish():
        raise PermissionDenied

    if not revision.submitted_for_moderation:
        messages.error(
            request,
            _("The page '{0}' is not currently awaiting moderation.").format(
                revision.page.specific_deferred.get_admin_display_title()
            ),
        )
        return redirect("wagtailadmin_home")

    if request.method == "POST":
        revision.reject_moderation(user=request.user)

        messages.success(
            request,
            _("Page '{0}' rejected for publication.").format(
                revision.page.specific_deferred.get_admin_display_title()
            ),
            buttons=[
                messages.button(
                    reverse("wagtailadmin_pages:edit", args=(revision.page.id,)),
                    _("Edit"),
                )
            ],
        )

        if not send_moderation_notification(revision, "rejected", request.user):
            messages.error(request, _("Failed to send rejection notifications"))

    return redirect("wagtailadmin_home")


@require_GET
def preview_for_moderation(request, revision_id):
    revision = get_object_or_404(PageRevision, id=revision_id)
    if not revision.page.permissions_for_user(request.user).can_publish():
        raise PermissionDenied

    if not revision.submitted_for_moderation:
        messages.error(
            request,
            _("The page '{0}' is not currently awaiting moderation.").format(
                revision.page.specific_deferred.get_admin_display_title()
            ),
        )
        return redirect("wagtailadmin_home")

    page = revision.as_page_object()

    try:
        preview_mode = page.default_preview_mode
    except IndexError:
        raise PermissionDenied

    return page.make_preview_request(
        request, preview_mode, extra_request_attrs={"revision_id": revision_id}
    )



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0013_update_golive_expire_help_text"),
    ]

    operations = [
        migrations.AlterField(
            model_name="grouppagepermission",
            name="group",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                verbose_name="Group",
                related_name="page_permissions",
                to="auth.Group",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="grouppagepermission",
            name="page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                verbose_name="Page",
                related_name="group_permissions",
                to="wagtailcore.Page",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="grouppagepermission",
            name="permission_type",
            field=models.CharField(
                choices=[
                    ("add", "Add/edit pages you own"),
                    ("edit", "Add/edit any page"),
                    ("publish", "Publish any page"),
                    ("lock", "Lock/unlock any page"),
                ],
                max_length=20,
                verbose_name="Permission type",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="page",
            name="search_description",
            field=models.TextField(blank=True, verbose_name="Search description"),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="page",
            name="show_in_menus",
            field=models.BooleanField(
                default=False,
                help_text="Whether a link to this page will appear in automatically generated menus",
                verbose_name="Show in menus",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="page",
            name="slug",
            field=models.SlugField(
                help_text="The name of the page as it will appear in URLs e.g http://domain.com/blog/[my-slug]/",
                max_length=255,
                verbose_name="Slug",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="page",
            name="title",
            field=models.CharField(
                help_text="The page title as you'd like it to be seen by the public",
                max_length=255,
                verbose_name="Title",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="pageviewrestriction",
            name="page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                verbose_name="Page",
                related_name="view_restrictions",
                to="wagtailcore.Page",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="pageviewrestriction",
            name="password",
            field=models.CharField(max_length=255, verbose_name="Password"),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="site",
            name="hostname",
            field=models.CharField(
                db_index=True, max_length=255, verbose_name="Hostname"
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="site",
            name="is_default_site",
            field=models.BooleanField(
                default=False,
                help_text="If true, this site will handle requests for all other hostnames"
                " that do not have a site entry of their own",
                verbose_name="Is default site",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="site",
            name="port",
            field=models.IntegerField(
                default=80,
                help_text="Set this to something other than 80 if you need a specific port number"
                " to appear in URLs (e.g. development on port 8000). Does not affect request handling"
                " (so port forwarding still works).",
                verbose_name="Port",
            ),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name="site",
            name="root_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                verbose_name="Root page",
                related_name="sites_rooted_here",
                to="wagtailcore.Page",
            ),
            preserve_default=True,
        ),
    ]



# Generated by Django 2.1.2 on 2018-10-28 11:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tests", "0039_customdocument_description"),
    ]

    operations = [
        migrations.AddField(
            model_name="customdocument",
            name="file_hash",
            field=models.CharField(blank=True, editable=False, max_length=40),
        ),
    ]



from django.urls import include, re_path

from wagtail import hooks


class ViewSetRegistry:
    def __init__(self):
        self.viewsets = []

    def populate(self):
        for fn in hooks.get_hooks("register_admin_viewset"):
            viewset = fn()
            self.register(viewset)

    def register(self, viewset):
        self.viewsets.append(viewset)
        viewset.on_register()
        return viewset

    def get_urlpatterns(self):
        urlpatterns = []

        for viewset in self.viewsets:
            vs_urlpatterns = viewset.get_urlpatterns()

            if vs_urlpatterns:
                urlpatterns.append(
                    re_path(
                        r"^{}/".format(viewset.url_prefix),
                        include((vs_urlpatterns, viewset.name), namespace=viewset.name),
                    )
                )

        return urlpatterns


viewsets = ViewSetRegistry()



from django.forms import widgets

from wagtail.telepath import register
from wagtail.widget_adapters import WidgetAdapter


class AdminAutoHeightTextInput(widgets.Textarea):
    template_name = "wagtailadmin/widgets/auto_height_text_input.html"

    def __init__(self, attrs=None):
        # Use more appropriate rows default, given autoheight will alter this anyway
        default_attrs = {"rows": "1"}
        if attrs:
            default_attrs.update(attrs)

        super().__init__(default_attrs)


class AdminAutoHeightTextInputAdapter(WidgetAdapter):
    js_constructor = "wagtail.widgets.AdminAutoHeightTextInput"


register(AdminAutoHeightTextInputAdapter(), AdminAutoHeightTextInput)



# -*- coding: utf-8 -*-
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailusers", "0003_add_verbose_names"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="userprofile",
            options={"verbose_name": "user profile"},
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="approved_notifications",
            field=models.BooleanField(
                default=True,
                help_text="Receive notification when your page edit is approved",
                verbose_name="approved notifications",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="rejected_notifications",
            field=models.BooleanField(
                default=True,
                help_text="Receive notification when your page edit is rejected",
                verbose_name="rejected notifications",
            ),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="submitted_notifications",
            field=models.BooleanField(
                default=True,
                help_text="Receive notification when a page is submitted for moderation",
                verbose_name="submitted notifications",
            ),
        ),
    ]



from datetime import date

from django.core.exceptions import ValidationError
from django.core.paginator import Paginator
from django.db import models
from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.fields import ParentalKey
from taggit.models import TaggedItemBase

from wagtail.admin.panels import FieldPanel, InlinePanel, MultiFieldPanel
from wagtail.api import APIField
from wagtail.fields import RichTextField
from wagtail.images.api.fields import ImageRenditionField
from wagtail.models import Orderable, Page
from wagtail.search import index

# ABSTRACT MODELS
# =============================


class AbstractLinkFields(models.Model):
    link_external = models.URLField("External link", blank=True)
    link_page = models.ForeignKey(
        "wagtailcore.Page",
        null=True,
        blank=True,
        related_name="+",
        on_delete=models.CASCADE,
    )
    link_document = models.ForeignKey(
        "wagtaildocs.Document",
        null=True,
        blank=True,
        related_name="+",
        on_delete=models.CASCADE,
    )

    @property
    def link(self):
        if self.link_page:
            return self.link_page.url
        elif self.link_document:
            return self.link_document.url
        else:
            return self.link_external

    def clean(self):
        if (
            self.link_page is None
            and self.link_document is None
            and not self.link_external
        ):
            raise ValidationError(
                "You must provide a related page, related document or an external URL"
            )

    api_fields = ("link",)

    panels = [
        FieldPanel("link_external"),
        FieldPanel("link_page"),
        FieldPanel("link_document"),
    ]

    class Meta:
        abstract = True


class AbstractRelatedLink(AbstractLinkFields):
    title = models.CharField(max_length=255, help_text="Link title")

    api_fields = ("title",) + AbstractLinkFields.api_fields

    panels = [
        FieldPanel("title"),
        MultiFieldPanel(AbstractLinkFields.panels, "Link"),
    ]

    class Meta:
        abstract = True


class AbstractCarouselItem(AbstractLinkFields):
    image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    embed_url = models.URLField("Embed URL", blank=True)
    caption = models.CharField(max_length=255, blank=True)

    api_fields = (
        "image",
        "embed_url",
        "caption",
    ) + AbstractLinkFields.api_fields

    panels = [
        FieldPanel("image"),
        FieldPanel("embed_url"),
        FieldPanel("caption"),
        MultiFieldPanel(AbstractLinkFields.panels, "Link"),
    ]

    class Meta:
        abstract = True


class ContactFieldsMixin(models.Model):
    telephone = models.CharField(max_length=20, blank=True)
    email = models.EmailField(blank=True)
    address_1 = models.CharField(max_length=255, blank=True)
    address_2 = models.CharField(max_length=255, blank=True)
    city = models.CharField(max_length=255, blank=True)
    country = models.CharField(max_length=255, blank=True)
    post_code = models.CharField(max_length=10, blank=True)

    api_fields = (
        "telephone",
        "email",
        "address_1",
        "address_2",
        "city",
        "country",
        "post_code",
    )

    panels = [
        FieldPanel("telephone"),
        FieldPanel("email"),
        FieldPanel("address_1"),
        FieldPanel("address_2"),
        FieldPanel("city"),
        FieldPanel("country"),
        FieldPanel("post_code"),
    ]

    class Meta:
        abstract = True


# PAGE MODELS
# =============================

# Home page


class HomePage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    body = RichTextField(blank=True)

    api_fields = (
        "body",
        "carousel_items",
        "related_links",
    )

    search_fields = Page.search_fields + [
        index.SearchField("body"),
    ]

    class Meta:
        verbose_name = "homepage"


class HomePageCarouselItem(Orderable, AbstractCarouselItem):
    page = ParentalKey(
        "HomePage", related_name="carousel_items", on_delete=models.CASCADE
    )


class HomePageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "HomePage", related_name="related_links", on_delete=models.CASCADE
    )


HomePage.content_panels = Page.content_panels + [
    FieldPanel("body", classname="full"),
    InlinePanel("carousel_items", label="Carousel items"),
    InlinePanel("related_links", label="Related links"),
]


# Standard pages


class StandardPage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    intro = RichTextField(blank=True)
    body = RichTextField(blank=True)
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        "intro",
        "body",
        "feed_image",
        "carousel_items",
        "related_links",
    )

    search_fields = Page.search_fields + [
        index.SearchField("intro"),
        index.SearchField("body"),
    ]


class StandardPageCarouselItem(Orderable, AbstractCarouselItem):
    page = ParentalKey(
        "StandardPage", related_name="carousel_items", on_delete=models.CASCADE
    )


class StandardPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "StandardPage", related_name="related_links", on_delete=models.CASCADE
    )


StandardPage.content_panels = Page.content_panels + [
    FieldPanel("intro", classname="full"),
    InlinePanel("carousel_items", label="Carousel items"),
    FieldPanel("body", classname="full"),
    InlinePanel("related_links", label="Related links"),
]


StandardPage.promote_panels = [
    MultiFieldPanel(Page.promote_panels, "Common page configuration"),
    FieldPanel("feed_image"),
]


class StandardIndexPage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    intro = RichTextField(blank=True)
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        "intro",
        "feed_image",
        "related_links",
    )

    search_fields = Page.search_fields + [
        index.SearchField("intro"),
    ]


class StandardIndexPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "StandardIndexPage", related_name="related_links", on_delete=models.CASCADE
    )


StandardIndexPage.content_panels = Page.content_panels + [
    FieldPanel("intro", classname="full"),
    InlinePanel("related_links", label="Related links"),
]


StandardIndexPage.promote_panels = [
    MultiFieldPanel(Page.promote_panels, "Common page configuration"),
    FieldPanel("feed_image"),
]


# Blog pages


class BlogEntryPage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    body = RichTextField()
    tags = ClusterTaggableManager(through="BlogEntryPageTag", blank=True)
    date = models.DateField("Post date")
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        APIField("body"),
        APIField("tags"),
        APIField("date"),
        APIField("feed_image"),
        APIField(
            "feed_image_thumbnail",
            serializer=ImageRenditionField("fill-300x300", source="feed_image"),
        ),
        APIField("carousel_items"),
        APIField("related_links"),
    )

    search_fields = Page.search_fields + [
        index.SearchField("body"),
    ]

    def get_blog_index(self):
        # Find closest ancestor which is a blog index
        return BlogIndexPage.ancestor_of(self).last()


class BlogEntryPageCarouselItem(Orderable, AbstractCarouselItem):
    page = ParentalKey(
        "BlogEntryPage", related_name="carousel_items", on_delete=models.CASCADE
    )


class BlogEntryPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "BlogEntryPage", related_name="related_links", on_delete=models.CASCADE
    )


class BlogEntryPageTag(TaggedItemBase):
    content_object = ParentalKey(
        "BlogEntryPage", related_name="tagged_items", on_delete=models.CASCADE
    )


BlogEntryPage.content_panels = Page.content_panels + [
    FieldPanel("date"),
    FieldPanel("body", classname="full"),
    InlinePanel("carousel_items", label="Carousel items"),
    InlinePanel("related_links", label="Related links"),
]


BlogEntryPage.promote_panels = [
    MultiFieldPanel(Page.promote_panels, "Common page configuration"),
    FieldPanel("feed_image"),
    FieldPanel("tags"),
]


class BlogIndexPage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    intro = RichTextField(blank=True)

    api_fields = (
        "intro",
        "related_links",
    )

    search_fields = Page.search_fields + [
        index.SearchField("intro"),
    ]

    def get_blog_entries(self):
        # Get list of live blog pages that are descendants of this page
        entries = BlogEntryPage.objects.descendant_of(self).live()

        # Order by most recent date first
        entries = entries.order_by("-date")

        return entries

    def get_context(self, request):
        # Get blog entries
        entries = self.get_blog_entries()

        # Filter by tag
        tag = request.GET.get("tag")
        if tag:
            entries = entries.filter(tags__name=tag)

        paginator = Paginator(entries, per_page=10)
        entries = paginator.get_page(request.GET.get("page"))

        # Update template context
        context = super().get_context(request)
        context["entries"] = entries
        return context


class BlogIndexPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "BlogIndexPage", related_name="related_links", on_delete=models.CASCADE
    )


BlogIndexPage.content_panels = Page.content_panels + [
    FieldPanel("intro", classname="full"),
    InlinePanel("related_links", label="Related links"),
]


# Events pages


class EventPage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    AUDIENCE_CHOICES = (
        ("public", "Public"),
        ("private", "Private"),
    )

    date_from = models.DateField("Start date")
    date_to = models.DateField(
        "End date",
        null=True,
        blank=True,
        help_text="Not required if event is on a single day",
    )
    time_from = models.TimeField("Start time", null=True, blank=True)
    time_to = models.TimeField("End time", null=True, blank=True)
    audience = models.CharField(max_length=255, choices=AUDIENCE_CHOICES)
    location = models.CharField(max_length=255)
    body = RichTextField(blank=True)
    cost = models.CharField(max_length=255)
    signup_link = models.URLField(blank=True)
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        "date_from",
        "date_to",
        "time_from",
        "time_to",
        "audience",
        "location",
        "body",
        "cost",
        "signup_link",
        "feed_image",
        "carousel_items",
        "related_links",
        "speakers",
    )

    search_fields = Page.search_fields + [
        index.SearchField("get_audience_display"),
        index.SearchField("location"),
        index.SearchField("body"),
    ]

    def get_event_index(self):
        # Find closest ancestor which is an event index
        return EventIndexPage.objects.ancester_of(self).last()


class EventPageCarouselItem(Orderable, AbstractCarouselItem):
    page = ParentalKey(
        "EventPage", related_name="carousel_items", on_delete=models.CASCADE
    )


class EventPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "EventPage", related_name="related_links", on_delete=models.CASCADE
    )


class EventPageSpeaker(Orderable, AbstractLinkFields):
    page = ParentalKey("EventPage", related_name="speakers", on_delete=models.CASCADE)
    first_name = models.CharField("Name", max_length=255, blank=True)
    last_name = models.CharField("Surname", max_length=255, blank=True)
    image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        "first_name",
        "last_name",
        "image",
    )

    panels = [
        FieldPanel("first_name"),
        FieldPanel("last_name"),
        FieldPanel("image"),
        MultiFieldPanel(AbstractLinkFields.panels, "Link"),
    ]


EventPage.content_panels = Page.content_panels + [
    FieldPanel("date_from"),
    FieldPanel("date_to"),
    FieldPanel("time_from"),
    FieldPanel("time_to"),
    FieldPanel("location"),
    FieldPanel("audience"),
    FieldPanel("cost"),
    FieldPanel("signup_link"),
    InlinePanel("carousel_items", label="Carousel items"),
    FieldPanel("body", classname="full"),
    InlinePanel("speakers", label="Speakers"),
    InlinePanel("related_links", label="Related links"),
]


EventPage.promote_panels = [
    MultiFieldPanel(Page.promote_panels, "Common page configuration"),
    FieldPanel("feed_image"),
]


class EventIndexPage(Page):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    intro = RichTextField(blank=True)

    api_fields = (
        "intro",
        "related_links",
    )

    search_fields = Page.search_fields + [
        index.SearchField("intro"),
    ]

    def get_events(self):
        # Get list of live event pages that are descendants of this page
        events = EventPage.objects.descendant_of(self).live()

        # Filter events list to get ones that are either
        # running now or start in the future
        events = events.filter(date_from__gte=date.today())

        # Order by date
        events = events.order_by("date_from")

        return events


class EventIndexPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "EventIndexPage", related_name="related_links", on_delete=models.CASCADE
    )


EventIndexPage.content_panels = Page.content_panels + [
    FieldPanel("intro", classname="full"),
    InlinePanel("related_links", label="Related links"),
]


# Person page


class PersonPage(Page, ContactFieldsMixin):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    first_name = models.CharField(max_length=255)
    last_name = models.CharField(max_length=255)
    intro = RichTextField(blank=True)
    biography = RichTextField(blank=True)
    image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        "first_name",
        "last_name",
        "intro",
        "biography",
        "image",
        "feed_image",
        "related_links",
    ) + ContactFieldsMixin.api_fields

    search_fields = Page.search_fields + [
        index.SearchField("first_name"),
        index.SearchField("last_name"),
        index.SearchField("intro"),
        index.SearchField("biography"),
    ]


class PersonPageRelatedLink(Orderable, AbstractRelatedLink):
    page = ParentalKey(
        "PersonPage", related_name="related_links", on_delete=models.CASCADE
    )


PersonPage.content_panels = Page.content_panels + [
    FieldPanel("first_name"),
    FieldPanel("last_name"),
    FieldPanel("intro", classname="full"),
    FieldPanel("biography", classname="full"),
    FieldPanel("image"),
    MultiFieldPanel(ContactFieldsMixin.panels, "Contact"),
    InlinePanel("related_links", label="Related links"),
]


PersonPage.promote_panels = [
    MultiFieldPanel(Page.promote_panels, "Common page configuration"),
    FieldPanel("feed_image"),
]


# Contact page


class ContactPage(Page, ContactFieldsMixin):
    page_ptr = models.OneToOneField(
        Page, parent_link=True, related_name="+", on_delete=models.CASCADE
    )
    body = RichTextField(blank=True)
    feed_image = models.ForeignKey(
        "wagtailimages.Image",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="+",
    )

    api_fields = (
        "body",
        "feed_image",
    ) + ContactFieldsMixin.api_fields

    search_fields = Page.search_fields + [
        index.SearchField("body"),
    ]


ContactPage.content_panels = Page.content_panels + [
    FieldPanel("body", classname="full"),
    MultiFieldPanel(ContactFieldsMixin.panels, "Contact"),
]


ContactPage.promote_panels = [
    MultiFieldPanel(Page.promote_panels, "Common page configuration"),
    FieldPanel("feed_image"),
]



from django.utils.http import url_has_allowed_host_and_scheme


def get_valid_next_url_from_request(request):
    next_url = request.POST.get("next") or request.GET.get("next")
    if not next_url or not url_has_allowed_host_and_scheme(
        url=next_url, allowed_hosts={request.get_host()}
    ):
        return ""
    return next_url



import json

from django import forms
from django.core.exceptions import ImproperlyConfigured
from django.forms import widgets
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from wagtail.admin.staticfiles import versioned_static
from wagtail.coreutils import resolve_model_string
from wagtail.models import Page
from wagtail.telepath import register
from wagtail.utils.widgets import WidgetWithScript
from wagtail.widget_adapters import WidgetAdapter


class AdminChooser(WidgetWithScript, widgets.Input):
    input_type = "hidden"
    choose_one_text = _("Choose an item")
    choose_another_text = _("Choose another item")
    clear_choice_text = _("Clear choice")
    link_to_chosen_text = _("Edit this item")
    show_edit_link = True
    show_clear_link = True

    # when looping over form fields, this one should appear in visible_fields, not hidden_fields
    # despite the underlying input being type="hidden"
    is_hidden = False

    def get_instance(self, model_class, value):
        # helper method for cleanly turning 'value' into an instance object.
        # DEPRECATED - subclasses should override WidgetWithScript.get_value_data instead
        if value is None:
            return None

        try:
            return model_class.objects.get(pk=value)
        except model_class.DoesNotExist:
            return None

    def get_instance_and_id(self, model_class, value):
        # DEPRECATED - subclasses should override WidgetWithScript.get_value_data instead
        if value is None:
            return (None, None)
        elif isinstance(value, model_class):
            return (value, value.pk)
        else:
            try:
                return (model_class.objects.get(pk=value), value)
            except model_class.DoesNotExist:
                return (None, None)

    def value_from_datadict(self, data, files, name):
        # treat the empty string as None
        result = super().value_from_datadict(data, files, name)
        if result == "":
            return None
        else:
            return result

    def __init__(self, **kwargs):
        # allow choose_one_text / choose_another_text to be overridden per-instance
        if "choose_one_text" in kwargs:
            self.choose_one_text = kwargs.pop("choose_one_text")
        if "choose_another_text" in kwargs:
            self.choose_another_text = kwargs.pop("choose_another_text")
        if "clear_choice_text" in kwargs:
            self.clear_choice_text = kwargs.pop("clear_choice_text")
        if "link_to_chosen_text" in kwargs:
            self.link_to_chosen_text = kwargs.pop("link_to_chosen_text")
        if "show_edit_link" in kwargs:
            self.show_edit_link = kwargs.pop("show_edit_link")
        if "show_clear_link" in kwargs:
            self.show_clear_link = kwargs.pop("show_clear_link")
        super().__init__(**kwargs)


class AdminPageChooser(AdminChooser):
    choose_one_text = _("Choose a page")
    choose_another_text = _("Choose another page")
    link_to_chosen_text = _("Edit this page")

    def __init__(
        self, target_models=None, can_choose_root=False, user_perms=None, **kwargs
    ):
        super().__init__(**kwargs)

        if target_models:
            if not isinstance(target_models, (set, list, tuple)):
                # assume we've been passed a single instance; wrap it as a list
                target_models = [target_models]

            # normalise the list of target models to a list of page classes
            cleaned_target_models = []
            for model in target_models:
                try:
                    cleaned_target_models.append(resolve_model_string(model))
                except (ValueError, LookupError):
                    raise ImproperlyConfigured(
                        "Could not resolve %r into a model. "
                        "Model names should be in the form app_label.model_name"
                        % (model,)
                    )
        else:
            cleaned_target_models = [Page]

        if len(cleaned_target_models) == 1 and cleaned_target_models[0] is not Page:
            model_name = cleaned_target_models[0]._meta.verbose_name.title()
            self.choose_one_text += " (" + model_name + ")"

        self.user_perms = user_perms
        self.target_models = cleaned_target_models
        self.can_choose_root = bool(can_choose_root)

    def _get_lowest_common_page_class(self):
        """
        Return a Page class that is an ancestor for all Page classes in
        ``target_models``, and is also a concrete Page class itself.
        """
        if len(self.target_models) == 1:
            # Shortcut for a single page type
            return self.target_models[0]
        else:
            return Page

    @property
    def model_names(self):
        return [
            "{app}.{model}".format(
                app=model._meta.app_label, model=model._meta.model_name
            )
            for model in self.target_models
        ]

    @property
    def client_options(self):
        # a JSON-serializable representation of the configuration options needed for the
        # client-side behaviour of this widget
        return {
            "model_names": self.model_names,
            "can_choose_root": self.can_choose_root,
            "user_perms": self.user_perms,
        }

    def get_value_data(self, value):
        if value is None:
            return None
        elif isinstance(value, Page):
            page = value.specific
        else:  # assume page ID
            model_class = self._get_lowest_common_page_class()
            try:
                page = model_class.objects.get(pk=value)
            except model_class.DoesNotExist:
                return None

            page = page.specific

        parent_page = page.get_parent()
        return {
            "id": page.pk,
            "display_title": page.get_admin_display_title(),
            "parent_id": parent_page.pk if parent_page else None,
            "edit_url": reverse("wagtailadmin_pages:edit", args=[page.pk]),
        }

    def render_html(self, name, value_data, attrs):
        value_data = value_data or {}
        original_field_html = super().render_html(name, value_data.get("id"), attrs)

        return render_to_string(
            "wagtailadmin/widgets/page_chooser.html",
            {
                "widget": self,
                "original_field_html": original_field_html,
                "attrs": attrs,
                "value": bool(
                    value_data
                ),  # only used by chooser.html to identify blank values
                "display_title": value_data.get("display_title", ""),
                "edit_url": value_data.get("edit_url", ""),
            },
        )

    def render_js_init(self, id_, name, value_data):
        value_data = value_data or {}
        return "createPageChooser({id}, {parent}, {options});".format(
            id=json.dumps(id_),
            parent=json.dumps(value_data.get("parent_id")),
            options=json.dumps(self.client_options),
        )

    @property
    def media(self):
        return forms.Media(
            js=[
                versioned_static("wagtailadmin/js/page-chooser-modal.js"),
                versioned_static("wagtailadmin/js/page-chooser.js"),
            ]
        )


class PageChooserAdapter(WidgetAdapter):
    js_constructor = "wagtail.widgets.PageChooser"

    def js_args(self, widget):
        return [
            widget.render_html("__NAME__", None, attrs={"id": "__ID__"}),
            widget.id_for_label("__ID__"),
            widget.client_options,
        ]


register(PageChooserAdapter(), AdminPageChooser)



from django.urls import path

from wagtail.users.views import users

app_name = "wagtailusers_users"
urlpatterns = [
    path("", users.index, name="index"),
    path("add/", users.create, name="add"),
    path("<str:user_id>/", users.edit, name="edit"),
    path("<str:user_id>/delete/", users.delete, name="delete"),
]



from functools import total_ordering

from django.forms.utils import flatatt
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.utils.html import format_html

from wagtail import hooks


@total_ordering
class Button:
    show = True

    def __init__(
        self, label, url, classes=set(), icon_name=None, attrs={}, priority=1000
    ):
        self.label = label
        self.url = url
        self.classes = classes
        self.icon_name = icon_name
        self.attrs = attrs.copy()
        self.priority = priority

    def render(self):
        attrs = {
            "href": self.url,
            "class": " ".join(sorted(self.classes)),
            "title": self.label,
        }
        attrs.update(self.attrs)
        return format_html("<a{}>{}</a>", flatatt(attrs), self.label)

    def __str__(self):
        return self.render()

    def __repr__(self):
        return "<Button: {}>".format(self.label)

    def __lt__(self, other):
        if not isinstance(other, Button):
            return NotImplemented
        return (self.priority, self.label) < (other.priority, other.label)

    def __eq__(self, other):
        if not isinstance(other, Button):
            return NotImplemented
        return (
            self.label == other.label
            and self.url == other.url
            and self.classes == other.classes
            and self.attrs == other.attrs
            and self.priority == other.priority
        )


# Base class for all listing buttons
# This is also used by SnippetListingButton defined in wagtail.snippets.widgets
class ListingButton(Button):
    def __init__(self, label, url, classes=set(), **kwargs):
        classes = {"button", "button-small", "button-secondary"} | set(classes)
        super().__init__(label, url, classes=classes, **kwargs)


class PageListingButton(ListingButton):
    pass


class BaseDropdownMenuButton(Button):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, url=None, **kwargs)

    @cached_property
    def dropdown_buttons(self):
        raise NotImplementedError

    def get_context_data(self):
        return {
            "buttons": self.dropdown_buttons,
            "label": self.label,
            "title": self.attrs.get("title"),
            "is_parent": self.is_parent,
        }

    def render(self):
        return render_to_string(self.template_name, self.get_context_data())


class ButtonWithDropdown(BaseDropdownMenuButton):
    template_name = "wagtailadmin/pages/listing/_button_with_dropdown.html"

    def __init__(self, *args, **kwargs):
        self.button_classes = kwargs.pop("button_classes", set())
        self.buttons_data = kwargs.pop("buttons_data", [])
        super().__init__(*args, **kwargs)

    def get_context_data(self):
        context = super().get_context_data()
        context["button_classes"] = self.button_classes
        context["classes"] = self.classes
        return context

    @cached_property
    def dropdown_buttons(self):
        return [Button(**button) for button in self.buttons_data]


class ButtonWithDropdownFromHook(BaseDropdownMenuButton):
    template_name = "wagtailadmin/pages/listing/_button_with_dropdown.html"

    def __init__(
        self, label, hook_name, page, page_perms, is_parent, next_url=None, **kwargs
    ):
        self.hook_name = hook_name
        self.page = page
        self.page_perms = page_perms
        self.is_parent = is_parent
        self.next_url = next_url

        super().__init__(label, **kwargs)

    @property
    def show(self):
        return bool(self.dropdown_buttons)

    @cached_property
    def dropdown_buttons(self):
        button_hooks = hooks.get_hooks(self.hook_name)

        buttons = []
        for hook in button_hooks:
            buttons.extend(
                hook(self.page, self.page_perms, self.is_parent, self.next_url)
            )

        buttons.sort()
        return buttons
